import io
from typing import Dict, Tuple

import aiohttp

from vocode import getenv
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import CoquiSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult
from vocode.streaming.utils.async_requester import AsyncRequestor

raise DeprecationWarning("This Synthesizer is deprecated and will be removed in the future.")

COQUI_BASE_URL = "https://app.coqui.ai/api/v2"


class CoquiSynthesizer(BaseSynthesizer[CoquiSynthesizerConfig]):
    def __init__(self, synthesizer_config: CoquiSynthesizerConfig):
        super().__init__(synthesizer_config)
        self.api_key = synthesizer_config.api_key or getenv("COQUI_API_KEY")
        self.voice_id = synthesizer_config.voice_id
        self.voice_prompt = synthesizer_config.voice_prompt
        self.use_xtts = synthesizer_config.use_xtts

    def get_request(self, text: str) -> Tuple[str, Dict[str, str], Dict[str, object]]:
        url = COQUI_BASE_URL
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        body = {
            "text": text,
            "speed": 1,
        }

        if self.use_xtts:
            # If we have a voice prompt, use that instead of the voice ID
            if self.voice_prompt is not None:
                url = f"{COQUI_BASE_URL}/samples/xtts/render-from-prompt/"
                body["prompt"] = self.voice_prompt
            else:
                url = f"{COQUI_BASE_URL}/samples/xtts/render/"
                body["voice_id"] = self.voice_id
        else:
            if self.voice_prompt is not None:
                url = f"{COQUI_BASE_URL}/samples/from-prompt/"
                body["prompt"] = self.voice_prompt
            else:
                url = f"{COQUI_BASE_URL}/samples"
                body["voice_id"] = self.voice_id
        return url, headers, body

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        url, headers, body = self.get_request(message.text)

        async with AsyncRequestor().get_session().request(
            "POST",
            url,
            json=body,
            headers=headers,
            timeout=aiohttp.ClientTimeout(total=15),
        ) as response:
            sample = await response.json()
            async with AsyncRequestor().get_session().request(
                "GET",
                sample["audio_url"],
            ) as response:
                read_response = await response.read()

                result = self.create_synthesis_result_from_wav(
                    synthesizer_config=self.synthesizer_config,
                    file=io.BytesIO(read_response),
                    message=message,
                    chunk_size=chunk_size,
                )
                return result



from enum import Enum
from typing import Optional

from .model import TypedModel

DEFAULT_EMBEDDINGS_MODEL = "text-embedding-ada-002"


class VectorDBType(str, Enum):
    BASE = "vector_db_base"
    PINECONE = "vector_db_pinecone"


class VectorDBConfig(TypedModel, type=VectorDBType.BASE.value):  # type: ignore
    embeddings_model: str = DEFAULT_EMBEDDINGS_MODEL


class PineconeConfig(VectorDBConfig, type=VectorDBType.PINECONE.value):  # type: ignore
    index: str
    api_key: Optional[str]
    api_environment: Optional[str]
    top_k: int = 3



from typing import Optional

from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.model import BaseModel


class InputAudioConfig(BaseModel):
    sampling_rate: int
    audio_encoding: AudioEncoding
    chunk_size: int
    downsampling: Optional[int] = None


class OutputAudioConfig(BaseModel):
    sampling_rate: int
    audio_encoding: AudioEncoding



from typing import Any, Dict, Set


def exclude_keys_recursive(d: Dict[str, Any], excluded_keys: Set[str]):
    if isinstance(d, dict):
        return {
            k: exclude_keys_recursive(v, excluded_keys)
            for k, v in d.items()
            if k not in excluded_keys
        }
    elif isinstance(d, list):
        return [exclude_keys_recursive(v, excluded_keys) for v in d]
    else:
        return d



import asyncio
import os
import re
from concurrent.futures import ThreadPoolExecutor
from typing import List, Optional
from xml.etree import ElementTree

import azure.cognitiveservices.speech as speechsdk
from loguru import logger

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding, SamplingRate
from vocode.streaming.models.message import BaseMessage, SSMLMessage
from vocode.streaming.models.synthesizer import AzureSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import (
    FILLER_AUDIO_PATH,
    FILLER_PHRASES,
    BaseSynthesizer,
    FillerAudio,
    SynthesisResult,
    encode_as_wav,
)

NAMESPACES = {
    "mstts": "https://www.w3.org/2001/mstts",
    "": "https://www.w3.org/2001/10/synthesis",
}

ElementTree.register_namespace("", NAMESPACES[""])
ElementTree.register_namespace("mstts", NAMESPACES["mstts"])

_AZURE_INSIDE_VOICE_REGEX = r"<voice[^>]*>(.*?)<\/voice>"


class WordBoundaryEventPool:
    def __init__(self):
        self.events = []

    def add(self, event):
        self.events.append(
            {
                "text": event.text,
                "text_offset": event.text_offset,
                "audio_offset": (event.audio_offset + 5000) / (10000 * 1000),
                "boudary_type": event.boundary_type,
            }
        )

    def get_events_sorted(self):
        return sorted(self.events, key=lambda event: event["audio_offset"])


class AzureSynthesizer(BaseSynthesizer[AzureSynthesizerConfig]):
    OFFSET_MS = 100

    def __init__(
        self,
        synthesizer_config: AzureSynthesizerConfig,
        azure_speech_key: Optional[str] = None,
        azure_speech_region: Optional[str] = None,
    ):
        super().__init__(synthesizer_config)
        # Instantiates a client
        azure_speech_key = azure_speech_key or getenv("AZURE_SPEECH_KEY")
        azure_speech_region = azure_speech_region or getenv("AZURE_SPEECH_REGION")
        if not azure_speech_key:
            raise ValueError(
                "Please set AZURE_SPEECH_KEY environment variable or pass it as a parameter"
            )
        if not azure_speech_region:
            raise ValueError(
                "Please set AZURE_SPEECH_REGION environment variable or pass it as a parameter"
            )
        speech_config = speechsdk.SpeechConfig(
            subscription=azure_speech_key, region=azure_speech_region
        )
        if self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            if self.synthesizer_config.sampling_rate == SamplingRate.RATE_44100:
                speech_config.set_speech_synthesis_output_format(
                    speechsdk.SpeechSynthesisOutputFormat.Raw44100Hz16BitMonoPcm
                )
            if self.synthesizer_config.sampling_rate == SamplingRate.RATE_48000:
                speech_config.set_speech_synthesis_output_format(
                    speechsdk.SpeechSynthesisOutputFormat.Raw48Khz16BitMonoPcm
                )
            if self.synthesizer_config.sampling_rate == SamplingRate.RATE_24000:
                speech_config.set_speech_synthesis_output_format(
                    speechsdk.SpeechSynthesisOutputFormat.Raw24Khz16BitMonoPcm
                )
            elif self.synthesizer_config.sampling_rate == SamplingRate.RATE_16000:
                speech_config.set_speech_synthesis_output_format(
                    speechsdk.SpeechSynthesisOutputFormat.Raw16Khz16BitMonoPcm
                )
            elif self.synthesizer_config.sampling_rate == SamplingRate.RATE_8000:
                speech_config.set_speech_synthesis_output_format(
                    speechsdk.SpeechSynthesisOutputFormat.Raw8Khz16BitMonoPcm
                )
        elif self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            speech_config.set_speech_synthesis_output_format(
                speechsdk.SpeechSynthesisOutputFormat.Raw8Khz8BitMonoMULaw
            )
        self.synthesizer = speechsdk.SpeechSynthesizer(
            speech_config=speech_config, audio_config=None
        )

        self.voice_name = self.synthesizer_config.voice_name
        self.pitch = self.synthesizer_config.pitch
        self.rate = self.synthesizer_config.rate
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: AzureSynthesizerConfig) -> str:
        return ":".join(
            (
                "azure",
                synthesizer_config.voice_name,
                str(synthesizer_config.pitch),
                str(synthesizer_config.rate),
                synthesizer_config.language_code,
                synthesizer_config.audio_encoding,
            )
        )

    async def get_phrase_filler_audios(self) -> List[FillerAudio]:
        filler_phrase_audios = []
        for filler_phrase in FILLER_PHRASES:
            cache_key = "-".join(
                (
                    str(filler_phrase.text),
                    str(self.synthesizer_config.type),
                    str(self.synthesizer_config.audio_encoding),
                    str(self.synthesizer_config.sampling_rate),
                    str(self.voice_name),
                    str(self.pitch),
                    str(self.rate),
                )
            )
            filler_audio_path = os.path.join(FILLER_AUDIO_PATH, f"{cache_key}.bytes")
            if os.path.exists(filler_audio_path):
                audio_data = open(filler_audio_path, "rb").read()
            else:
                logger.debug(f"Generating filler audio for {filler_phrase.text}")
                ssml = self.create_ssml(filler_phrase.text)
                result = await asyncio.get_event_loop().run_in_executor(
                    self.thread_pool_executor, self.synthesizer.speak_ssml, ssml
                )
                offset = self.synthesizer_config.sampling_rate * self.OFFSET_MS // 1000
                audio_data = result.audio_data[offset:]
                with open(filler_audio_path, "wb") as f:
                    f.write(audio_data)
            filler_phrase_audios.append(
                FillerAudio(
                    filler_phrase,
                    audio_data,
                    self.synthesizer_config,
                )
            )
        return filler_phrase_audios

    def add_marks(self, message: str, index=0) -> str:
        search_result = re.search(r"([\.\,\:\;\-\—]+)", message)
        if search_result is None:
            return message
        start, end = search_result.span()
        with_mark = message[:start] + f'<mark name="{index}" />' + message[start:end]
        rest = message[end:]
        rest_stripped = re.sub(r"^(.+)([\.\,\:\;\-\—]+)$", r"\1", rest)
        if len(rest_stripped) == 0:
            return with_mark
        return with_mark + self.add_marks(rest_stripped, index + 1)

    def word_boundary_cb(self, evt, pool):
        pool.add(evt)

    def create_ssml(self, message: str) -> str:
        voice_language_code = self.synthesizer_config.voice_name[:5]
        ssml_root = ElementTree.fromstring(
            f'<speak version="1.0" xmlns="https://www.w3.org/2001/10/synthesis" xml:lang="{voice_language_code}"></speak>'
        )
        voice = ElementTree.SubElement(ssml_root, "voice")
        voice.set("name", self.voice_name)
        if self.synthesizer_config.language_code != "en-US":
            lang = ElementTree.SubElement(voice, "{%s}lang" % NAMESPACES.get(""))
            lang.set("xml:lang", self.synthesizer_config.language_code)
            voice_root = lang
        else:
            voice_root = voice
        # this ugly hack is necessary so we can limit the gap between sentences
        # for normal sentences, it seems like the gap is > 500ms, so we're able to reduce it to 500ms
        # for very tiny sentences, the API hangs - so we heuristically only update the silence gap
        # if there is more than one word in the sentence
        if " " in message:
            silence = ElementTree.SubElement(voice_root, "{%s}silence" % NAMESPACES.get("mstts"))
            silence.set("value", "500ms")
            silence.set("type", "Tailing-exact")
        prosody = ElementTree.SubElement(voice_root, "prosody")
        prosody.set("pitch", f"{self.pitch}%")
        prosody.set("rate", f"{self.rate}%")
        prosody.text = message.strip()
        ssml = ElementTree.tostring(ssml_root, encoding="unicode")
        regmatch = re.search(_AZURE_INSIDE_VOICE_REGEX, ssml, re.DOTALL)
        if regmatch:
            self.total_chars += len(regmatch.group(1))
        return ssml

    def synthesize_ssml(self, ssml: str) -> speechsdk.AudioDataStream:
        result = self.synthesizer.start_speaking_ssml_async(ssml).get()
        return speechsdk.AudioDataStream(result)

    def ready_synthesizer(self, chunk_size: int):
        # TODO: remove warming up the synthesizer for now
        # connection = speechsdk.Connection.from_speech_synthesizer(self.synthesizer)
        # connection.open(True)
        pass

    # given the number of seconds the message was allowed to go until, where did we get in the message?
    def get_message_up_to(
        self,
        message: str,
        ssml: str,
        seconds: float,
        word_boundary_event_pool: WordBoundaryEventPool,
    ) -> str:
        events = word_boundary_event_pool.get_events_sorted()
        for event in events:
            if event["audio_offset"] > seconds:
                ssml_fragment = ssml[: event["text_offset"]]
                # TODO: this is a little hacky, but it works for now
                return ssml_fragment.split(">")[-1]
        return message

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        # offset = int(self.OFFSET_MS * (self.synthesizer_config.sampling_rate / 1000))
        offset = 0
        logger.debug(f"Synthesizing message: {message}")

        # Azure will return no audio for certain strings like "-", "[-", and "!"
        # which causes the `chunk_generator` below to hang. Return an empty
        # generator for these cases.
        if not re.search(r"\w", message.text):
            return SynthesisResult(
                self.empty_generator(),
                lambda _: message.text,
            )

        async def chunk_generator(
            audio_data_stream: speechsdk.AudioDataStream, chunk_transform=lambda x: x
        ):
            audio_buffer = bytes(chunk_size)
            filled_size = await asyncio.get_event_loop().run_in_executor(
                self.thread_pool_executor,
                lambda: audio_data_stream.read_data(audio_buffer),
            )
            if filled_size != chunk_size:
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer[offset:]), True)
                return
            else:
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer[offset:]), False)
            while True:
                audio_buffer = bytes(chunk_size)
                filled_size = audio_data_stream.read_data(audio_buffer)
                if filled_size != chunk_size:
                    yield SynthesisResult.ChunkResult(
                        chunk_transform(audio_buffer[: filled_size - offset]), True
                    )
                    break
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer), False)

        word_boundary_event_pool = WordBoundaryEventPool()
        self.synthesizer.synthesis_word_boundary.connect(
            lambda event: self.word_boundary_cb(event, word_boundary_event_pool)
        )
        ssml = message.ssml if isinstance(message, SSMLMessage) else self.create_ssml(message.text)
        audio_data_stream = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor, self.synthesize_ssml, ssml
        )
        if self.synthesizer_config.should_encode_as_wav:
            output_generator = chunk_generator(
                audio_data_stream,
                lambda chunk: encode_as_wav(chunk, self.synthesizer_config),
            )
        else:
            output_generator = chunk_generator(audio_data_stream)

        return SynthesisResult(
            output_generator,
            lambda seconds: self.get_message_up_to(
                message.text, ssml, seconds, word_boundary_event_pool
            ),
        )



import asyncio
import json
from datetime import datetime, timezone
from typing import List, Optional, Tuple, Union
from urllib.parse import urlencode

import sentry_sdk
import websockets
from loguru import logger
from pydantic.v1 import BaseModel, Field
from websockets.client import WebSocketClientProtocol

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import (
    DEEPGRAM_API_WS_URL,
    DeepgramTranscriberConfig,
    EndpointingConfig,
    PunctuationEndpointingConfig,
    TimeEndpointingConfig,
    Transcription,
)
from vocode.streaming.transcriber.base_transcriber import BaseAsyncTranscriber
from vocode.utils.sentry_utils import CustomSentrySpans, sentry_configured, sentry_create_span

PUNCTUATION_TERMINATORS = [".", "!", "?"]
NUM_RESTARTS = 5
NUM_AUDIO_CHANNELS = 1


def now():
    return datetime.now(tz=timezone.utc)


class TimeSilentConfig(BaseModel):
    time_cutoff_seconds: float = 1
    post_punctuation_time_seconds: float = 0.5


class InternalPunctuationEndpointingConfig(  # type: ignore
    EndpointingConfig, type="internal_punctuation_based"
):
    time_silent_config: TimeSilentConfig = Field(default_factory=TimeSilentConfig)
    use_single_utterance_endpointing_for_first_utterance: bool = False


class DeepgramEndpointingConfig(EndpointingConfig, type="deepgram"):  # type: ignore
    vad_threshold_ms: int = 500
    utterance_cutoff_ms: int = 1000
    time_silent_config: Optional[TimeSilentConfig] = Field(default_factory=TimeSilentConfig)
    use_single_utterance_endpointing_for_first_utterance: bool = False


class DeepgramUtteranceEnd(BaseModel):
    pass

    def __str__(self):
        return "DeepgramUtteranceEnd()"


class DeepgramTranscriptionResult(BaseModel):
    class TranscriptionChoice(BaseModel):
        transcript: str
        confidence: float
        words: List[dict]

    is_final: bool
    speech_final: bool
    top_choice: TranscriptionChoice
    start: float
    duration: float

    def __str__(self):
        return f"DeepgramTranscriptionResult(transcript={self.top_choice.transcript}, is_final={self.is_final}, speech_final={self.speech_final})"


class DeepgramTranscriber(BaseAsyncTranscriber[DeepgramTranscriberConfig]):
    def __init__(
        self,
        transcriber_config: DeepgramTranscriberConfig,
    ):
        super().__init__(transcriber_config)
        self.api_key = self.transcriber_config.api_key or getenv("DEEPGRAM_API_KEY")
        if not self.api_key:
            raise ValueError(
                "Please set DEEPGRAM_API_KEY environment variable or pass it as a parameter"
            )
        self._ended = False
        self.is_ready = False
        self.audio_cursor = 0.0
        self.transcript_cursor = 0.0

        self.ws_url = getattr(transcriber_config, "ws_url", DEEPGRAM_API_WS_URL)

        self.min_latency = 0.0
        self.max_latency = 0.0
        self.avg_latency = 0.0

        self.avg_latency_numer = 0.0
        self.avg_latency_denom = 0.0

        self.start_ts: Optional[datetime] = None
        self.connected_ts: Optional[datetime] = None
        self.start_sending_ts: Optional[datetime] = None
        self.start_receiving_ts: Optional[datetime] = None

        self.is_first_transcription = True

    def _get_speed_coefficient(self):
        return self.speed_manager.get_speed_coefficient() if self.speed_manager else 1.0

    def get_deepgram_url(self):
        if self.transcriber_config.audio_encoding == AudioEncoding.LINEAR16:
            encoding = "linear16"
        elif self.transcriber_config.audio_encoding == AudioEncoding.MULAW:
            encoding = "mulaw"
        else:
            raise Exception(
                f"Audio encoding not supported {self.transcriber_config.audio_encoding}"
            )
        url_params = {
            "encoding": encoding,
            "sample_rate": self.transcriber_config.sampling_rate,
            "channels": NUM_AUDIO_CHANNELS,
            "interim_results": "true",
        }
        extra_params = {}
        if self.transcriber_config.language:
            extra_params["language"] = self.transcriber_config.language
        if self.transcriber_config.model:
            extra_params["model"] = self.transcriber_config.model
        if self.transcriber_config.tier:
            extra_params["tier"] = self.transcriber_config.tier
        if self.transcriber_config.version:
            extra_params["version"] = self.transcriber_config.version
        if self.transcriber_config.keywords:
            extra_params["keywords"] = self.transcriber_config.keywords
        if self.transcriber_config.endpointing_config and (
            isinstance(
                self.transcriber_config.endpointing_config,
                PunctuationEndpointingConfig,
            )
            or isinstance(
                self.transcriber_config.endpointing_config,
                InternalPunctuationEndpointingConfig,
            )
            or isinstance(
                self.transcriber_config.endpointing_config,
                DeepgramEndpointingConfig,
            )
        ):
            extra_params["punctuate"] = "true"
        if isinstance(
            self.transcriber_config.endpointing_config,
            DeepgramEndpointingConfig,
        ):
            extra_params["endpointing"] = int(
                self.transcriber_config.endpointing_config.vad_threshold_ms
                * (1 / self._get_speed_coefficient())
            )
            extra_params["utterance_end_ms"] = max(
                int(
                    self.transcriber_config.endpointing_config.utterance_cutoff_ms
                    * (1 / self._get_speed_coefficient())
                ),
                1000,
            )  # Deepgram recommends using at least 1000ms since the tick period is ~1s
        url_params.update(extra_params)
        return f"{self.ws_url}/v1/listen?{urlencode(url_params)}"

    async def _run_loop(self):
        restarts = 0
        while not self._ended and restarts < NUM_RESTARTS:
            await self.process()
            restarts += 1
            logger.debug(f"Deepgram connection died, restarting, num_restarts: {restarts}")

        logger.error("Deepgram connection died, not restarting")

    def terminate(self):
        self._track_latency_of_transcription_start()
        # Put this in logs until we sentry metrics show up
        # properly on dashboard
        logger.info(
            f"Transcription latency is {self.avg_latency:.4f}s",
            extra={
                "avg_latency": self.avg_latency,
                "min_latency": self.min_latency,
                "max_latency": self.max_latency,
            },
        )
        terminate_msg = json.dumps({"type": "CloseStream"}).encode("utf-8")
        self.input_queue.put_nowait(terminate_msg)
        self._ended = True
        super().terminate()

    def get_input_sample_width(self):
        encoding = self.transcriber_config.audio_encoding
        if encoding == AudioEncoding.LINEAR16:
            return 2
        elif encoding == AudioEncoding.MULAW:
            return 1
        else:
            raise Exception(f"Audio encoding not supported {encoding}")

    def get_byte_rate(self):
        sample_width = self.get_input_sample_width()
        sample_rate = self.transcriber_config.sampling_rate
        return sample_width * sample_rate * NUM_AUDIO_CHANNELS

    def is_endpoint(
        self,
        current_buffer: str,
        deepgram_response: Union[DeepgramUtteranceEnd, DeepgramTranscriptionResult],
        time_silent: float,
    ):
        is_endpoint, log_params = self._compute_is_endpoint_and_log_params(
            current_buffer, deepgram_response, time_silent
        )
        if is_endpoint:
            logger.info("Endpoint detected", extra=log_params)
        return is_endpoint

    @staticmethod
    def _satisfies_time_cutoff(
        *,
        seconds: float,
        deepgram_response: Union[DeepgramUtteranceEnd, DeepgramTranscriptionResult],
        current_buffer: str,
        time_silent: float,
    ):
        return (
            isinstance(deepgram_response, DeepgramTranscriptionResult)
            and not deepgram_response.top_choice.transcript
            and len(current_buffer) > 0
            and (time_silent + deepgram_response.duration) > seconds
        )

    def _compute_is_endpoint_and_update_log_params_based_on_time_silent(
        self,
        current_buffer: str,
        deepgram_response: DeepgramTranscriptionResult,
        time_silent: float,
        time_silent_config: TimeSilentConfig,
        existing_log_params: dict,
    ) -> bool:
        if current_buffer.strip():
            if current_buffer.strip()[
                -1
            ] in PUNCTUATION_TERMINATORS and self._satisfies_time_cutoff(
                seconds=time_silent_config.post_punctuation_time_seconds
                * (1.0 / self._get_speed_coefficient()),
                deepgram_response=deepgram_response,
                current_buffer=current_buffer,
                time_silent=time_silent,
            ):
                existing_log_params["source"] = "punctuation"
                return True
            elif self._satisfies_time_cutoff(
                seconds=time_silent_config.time_cutoff_seconds
                * (1.0 / self._get_speed_coefficient()),
                deepgram_response=deepgram_response,
                current_buffer=current_buffer,
                time_silent=time_silent,
            ):
                existing_log_params["source"] = "time_cutoff"
                return True
        return False

    def _compute_is_endpoint_and_log_params(
        self,
        current_buffer: str,
        deepgram_response: Union[DeepgramUtteranceEnd, DeepgramTranscriptionResult],
        time_silent: float,
    ) -> Tuple[bool, dict]:
        endpointing_config = self.transcriber_config.endpointing_config
        assert endpointing_config is not None

        log_params = {"endpointing_type": endpointing_config.type}

        if isinstance(
            endpointing_config,
            TimeEndpointingConfig,
        ):
            # if it is time based, then return true if there is no transcript
            # and there is some speech to send
            # and the time_silent is greater than the cutoff
            return (
                self._satisfies_time_cutoff(
                    seconds=endpointing_config.time_cutoff_seconds,
                    deepgram_response=deepgram_response,
                    current_buffer=current_buffer,
                    time_silent=time_silent,
                ),
                log_params,
            )
        elif isinstance(
            endpointing_config,
            DeepgramEndpointingConfig,
        ):
            if (
                isinstance(deepgram_response, DeepgramTranscriptionResult)
                and self.is_first_transcription
                and endpointing_config.use_single_utterance_endpointing_for_first_utterance
            ):
                if (
                    deepgram_response.top_choice.transcript
                    and deepgram_response.top_choice.transcript.strip()[-1]
                    in PUNCTUATION_TERMINATORS
                ):
                    log_params["source"] = "is_final"
                    return True, log_params
            if isinstance(deepgram_response, DeepgramUtteranceEnd):
                log_params["source"] = "utterance_end"
                return True, log_params
            elif isinstance(deepgram_response, DeepgramTranscriptionResult):
                if deepgram_response.top_choice.transcript and deepgram_response.speech_final:
                    log_params["source"] = "speech_final"
                    return True, log_params
                elif (
                    endpointing_config.time_silent_config is not None
                    and self._compute_is_endpoint_and_update_log_params_based_on_time_silent(
                        current_buffer,
                        deepgram_response,
                        time_silent,
                        endpointing_config.time_silent_config,
                        log_params,
                    )
                ):
                    return True, log_params
            return False, log_params

        if isinstance(deepgram_response, DeepgramTranscriptionResult):
            if isinstance(
                endpointing_config,
                PunctuationEndpointingConfig,
            ):
                if (
                    deepgram_response.top_choice.transcript
                    and deepgram_response.speech_final
                    and deepgram_response.top_choice.transcript.strip()[-1]
                    in PUNCTUATION_TERMINATORS
                ):
                    log_params["source"] = "punctuation"
                    return True, log_params
                elif self._satisfies_time_cutoff(
                    seconds=endpointing_config.time_cutoff_seconds
                    * (1.0 / self._get_speed_coefficient()),
                    deepgram_response=deepgram_response,
                    current_buffer=current_buffer,
                    time_silent=time_silent,
                ):
                    log_params["source"] = "time_cutoff"
                    return True, log_params
            elif isinstance(
                endpointing_config,
                InternalPunctuationEndpointingConfig,
            ):
                if (
                    self.is_first_transcription
                    and endpointing_config.use_single_utterance_endpointing_for_first_utterance
                ):
                    if deepgram_response.top_choice.transcript and deepgram_response.speech_final:
                        log_params["source"] = "speech_final"
                        return True, log_params
                if self._compute_is_endpoint_and_update_log_params_based_on_time_silent(
                    current_buffer,
                    deepgram_response,
                    time_silent,
                    endpointing_config.time_silent_config,
                    log_params,
                ):
                    return True, log_params
        return False, log_params

    def calculate_time_silent(self, deepgram_transcription_result: DeepgramTranscriptionResult):
        end = deepgram_transcription_result.start + deepgram_transcription_result.duration
        words = deepgram_transcription_result.top_choice.words
        if words:
            return end - words[-1]["end"]
        return deepgram_transcription_result.duration

    def calculate_duration(self, words: List[dict]) -> float:
        if len(words) == 0:
            return 0.0
        return words[-1]["end"] - words[0]["start"]

    async def process(self):
        self.audio_cursor = 0.0
        self.start_ts = now()

        extra_headers = {"Authorization": f"Token {self.api_key}"}
        deepgram_url = self.get_deepgram_url()
        logger.info(f"Connecting to Deepgram at {deepgram_url}")

        try:
            async with websockets.connect(deepgram_url, extra_headers=extra_headers) as ws:
                self.connected_ts = now()

                async def sender(
                    ws: WebSocketClientProtocol,
                ):  # sends audio to websocket
                    byte_rate = self.get_byte_rate()

                    while not self._ended:
                        try:
                            data = await asyncio.wait_for(self.input_queue.get(), 5)
                        except asyncio.exceptions.TimeoutError:
                            break

                        self.audio_cursor += len(data) / byte_rate

                        if not self.start_sending_ts:
                            self.start_sending_ts = now()

                        await ws.send(data)

                    logger.debug("Terminating Deepgram transcriber sender")

                async def receiver(ws: WebSocketClientProtocol):
                    buffer = ""
                    buffer_avg_confidence = 0.0
                    num_buffer_utterances = 1
                    time_silent = 0.0
                    words_buffer = []
                    is_final_ts: Optional[datetime] = None

                    while not self._ended:
                        try:
                            msg = await ws.recv()
                            if not self.start_receiving_ts:
                                self.start_receiving_ts = now()
                        except Exception as e:
                            logger.debug(f"Got error {e} in Deepgram receiver")
                            break
                        data = json.loads(msg)

                        if "start" in data and "duration" in data:
                            self._track_transcription_latency(
                                start=data["start"],
                                duration=data["duration"],
                            )

                        deepgram_response: Union[DeepgramUtteranceEnd, DeepgramTranscriptionResult]

                        if data["type"] == "Results":
                            deepgram_response = DeepgramTranscriptionResult(
                                is_final=data["is_final"],
                                speech_final=data["speech_final"],
                                top_choice=data["channel"]["alternatives"][0],
                                duration=data["duration"],
                                start=data["start"],
                            )
                        elif data["type"] == "UtteranceEnd":
                            deepgram_response = DeepgramUtteranceEnd()
                        else:
                            logger.info(f"Ignoring deepgram response type: {data['type']}")
                            continue

                        if (
                            isinstance(deepgram_response, DeepgramTranscriptionResult)
                            and deepgram_response.top_choice.transcript
                            and deepgram_response.top_choice.confidence > 0.0
                            and deepgram_response.is_final
                        ):
                            words = deepgram_response.top_choice.words
                            if words:
                                words_buffer.extend(words)
                            buffer = f"{buffer} {deepgram_response.top_choice.transcript}"
                            if buffer_avg_confidence == 0:
                                buffer_avg_confidence = deepgram_response.top_choice.confidence
                            else:
                                buffer_avg_confidence = (
                                    buffer_avg_confidence
                                    + deepgram_response.top_choice.confidence
                                    / (num_buffer_utterances)
                                ) * (num_buffer_utterances / (num_buffer_utterances + 1))
                            num_buffer_utterances += 1

                            is_final_ts = now()

                        if buffer and self.is_endpoint(buffer, deepgram_response, time_silent):
                            output_ts = now()
                            self._track_latency_of_conversation(
                                is_final_ts=is_final_ts,
                                output_ts=output_ts,
                            )
                            self.output_queue.put_nowait(
                                Transcription(
                                    message=buffer,
                                    confidence=buffer_avg_confidence,
                                    is_final=True,
                                    duration_seconds=self.calculate_duration(words_buffer),
                                )
                            )
                            buffer = ""
                            buffer_avg_confidence = 0.0
                            num_buffer_utterances = 1
                            time_silent = 0.0
                            words_buffer = []
                            is_final_ts = None

                        if isinstance(deepgram_response, DeepgramTranscriptionResult):
                            if (
                                isinstance(deepgram_response, DeepgramTranscriptionResult)
                                and deepgram_response.top_choice.transcript
                                and deepgram_response.top_choice.confidence > 0.0
                            ):
                                if not deepgram_response.is_final:
                                    interim_message = (
                                        f"{buffer} {deepgram_response.top_choice.transcript}"
                                    )
                                else:
                                    interim_message = buffer

                                self.output_queue.put_nowait(
                                    Transcription(
                                        message=interim_message,
                                        confidence=deepgram_response.top_choice.confidence,
                                        is_final=False,
                                    )
                                )
                                time_silent = self.calculate_time_silent(deepgram_response)
                            else:
                                time_silent += deepgram_response.duration

                    logger.debug("Terminating Deepgram transcriber receiver")

                await asyncio.gather(sender(ws), receiver(ws))

        except asyncio.exceptions.TimeoutError:
            raise

    @sentry_configured
    def _track_latency_of_transcription_start(
        self,
    ):
        with sentry_create_span(
            sentry_callable=sentry_sdk.start_span,
            op=CustomSentrySpans.LATENCY_OF_TRANSCRIPTION_START,
            start_timestamp=self.start_ts,
        ) as transcription_span:

            with sentry_create_span(
                sentry_callable=transcription_span.start_child,
                op=CustomSentrySpans.START_TO_CONNECTION,
                start_timestamp=self.start_ts,
            ) as span:
                span.finish(end_timestamp=self.connected_ts)

            with sentry_create_span(
                sentry_callable=transcription_span.start_child,
                op=CustomSentrySpans.CONNECTED_TO_FIRST_SEND,
                start_timestamp=self.connected_ts,
            ) as span:
                span.finish(end_timestamp=self.start_sending_ts)

            with sentry_create_span(
                sentry_callable=transcription_span.start_child,
                op=CustomSentrySpans.FIRST_SEND_TO_FIRST_RECEIVE,
                start_timestamp=self.start_sending_ts,
            ) as span:
                span.finish(end_timestamp=self.start_receiving_ts)
            transcription_span.finish(end_timestamp=self.start_receiving_ts)

    def _track_transcription_latency(self, start: float, duration: float):
        cur_max_latency = self.audio_cursor - self.transcript_cursor
        transcript_cursor = start + duration
        cur_min_latency = self.audio_cursor - transcript_cursor

        self.max_latency = max(self.max_latency or 0, cur_max_latency)
        self.min_latency = min(self.min_latency or 10**6, cur_min_latency)

        self.avg_latency_numer += (cur_min_latency + cur_max_latency) / 2 * duration
        self.avg_latency_denom += duration
        self.avg_latency = self.avg_latency_numer / (self.avg_latency_denom or 1)

    def _track_latency_of_conversation(
        self,
        *,
        # is_final_spoken_ts: Optional[datetime],
        is_final_ts: Optional[datetime],
        output_ts: datetime,
    ):
        if not (
            output_ts
            and is_final_ts
            # and is_final_spoken_ts
        ):
            logger.info("Skipping latency measurement when not all timestamps are set")
            return

        if not (output_ts > is_final_ts):  # > is_final_spoken_ts
            logger.error("Skipping latency measurement when timestamps are not in order")
            return
        logger.debug(
            f"Endpoint detected, tracking endpointing_latency={output_ts - is_final_ts}",
        )

        # Ideally, we want to start a latency measurement for the entire conversation delay:
        # from when the user stopped speaking, to the time we start sending audio back.
        # Currently we don't have a way to find from Deepgram what the timetamp is
        # corresponding to their 0.0 in the transcript_cursor and we have observed that
        # the audio_cursor is always less than the transcript cursor, which seems to
        # indicate there is something off in our math for the audio cursor. The audio cursor
        # should represent the total number of seconds of audio that we have submitted to
        # Deepgram and should roughly line up with the transcript cursor.
        #
        # Until we can find a reliable way to find is_final_spoken_ts, we are skipping on
        # sending the transcription_latency span, instead focus on the endpointing_latency
        #
        # There was previous work done using the last word "end" cursor offset from Deepgram
        # in [EPD-926]. However, for similar reasons the logic for adding that to the
        # start_receiving_ts created on our end resulted in times for that audio that were
        # later than "is_final" which does not make sense temporally. We believe that is due
        # to not having an accurate timestamp corresponding to deepgram's 0.0 audio_cursor offset.

        # Span: latency_of_conversation

        # Once we have reliable way to find is_final_spoken_ts we would use that
        # here instead, and uncomment the transcription_latency span
        # start_timestamp=is_final_spoken_ts,
        latency_of_conversation_span = sentry_create_span(
            sentry_callable=sentry_sdk.start_span,
            op=CustomSentrySpans.LATENCY_OF_CONVERSATION,
            start_timestamp=is_final_ts,
        )

        # Span: latency_of_conversation.transcription_latency
        # Transcription latency is the time between audio was sent to deepgram for the
        # last word in the buffer when we set `is_final_ts` and then start
        # running
        # Time from `is_final_spoken_ts` to `is_final_ts`
        # with sentry_create_span(
        #     sentry_callable=latency_of_conversation_span.start_child,
        #     op="transcription_latency",
        #     start_timestamp=is_final_spoken_ts,
        # ) as span:
        #     span.finish(end_timestamp=is_final_ts)

        # Span: latency_of_conversation.endpointing_latency
        # Endpointing latency is the time from when we receive the last is_final
        # message from Deepgram and when we send output to the next stage.
        # Time from `is_final_ts` to `output_ts`
        if latency_of_conversation_span:
            with sentry_create_span(
                sentry_callable=latency_of_conversation_span.start_child,
                op=CustomSentrySpans.ENDPOINTING_LATENCY,
                start_timestamp=is_final_ts,
            ) as span:
                span.finish(end_timestamp=output_ts)



from typing import Optional, Tuple

from vocode.streaming.agent.base_agent import RespondAgent
from vocode.streaming.models.agent import GPT4AllAgentConfig
from vocode.turn_based.agent.gpt4all_agent import GPT4AllAgent as TurnBasedGPT4AllAgent

raise DeprecationWarning("This Agent is deprecated and will be removed in the future.")


class GPT4AllAgent(RespondAgent[GPT4AllAgentConfig]):
    def __init__(self, agent_config: GPT4AllAgentConfig):
        super().__init__(agent_config=agent_config)
        self.turn_based_agent = TurnBasedGPT4AllAgent(
            model_path=agent_config.model_path,
            system_prompt=agent_config.prompt_preamble,
            initial_message=(
                agent_config.initial_message.text if agent_config.initial_message else None
            ),
        )

    async def respond(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
    ) -> Tuple[Optional[str], bool]:
        return (await self.turn_based_agent.respond_async(human_input)), False



import asyncio
import io
import re
import typing
from typing import List, Optional

import aiohttp
import requests
from pydub import AudioSegment

from vocode import getenv
from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer

raise DeprecationWarning("This Synthesizer is deprecated and will be removed in the future.")

COQUI_BASE_URL = "https://app.coqui.ai/api/v2/samples"
DEFAULT_SPEAKER_ID = "d2bd7ccb-1b65-4005-9578-32c4e02d8ddf"
MAX_TEXT_LENGTH = 250  # The maximum length of text that can be synthesized at once


class CoquiSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        voice_id: Optional[str] = None,
        voice_prompt: Optional[str] = None,
        use_xtts: bool = False,
        api_key: Optional[str] = None,
    ):
        self.voice_id = voice_id or DEFAULT_SPEAKER_ID
        self.voice_prompt = voice_prompt
        self.use_xtts = use_xtts
        self.api_key = getenv("COQUI_API_KEY", api_key)

    def synthesize(self, text: str) -> AudioSegment:
        text_chunks = self.split_text(text)
        audio_chunks = [self.synthesize_chunk(chunk) for chunk in text_chunks]
        return sum(audio_chunks)  # type: ignore

    def synthesize_chunk(self, text: str) -> AudioSegment:
        url, headers, body = self.get_request(text)

        # Get the sample
        response = requests.post(url, headers=headers, json=body)
        assert response.ok, response.text
        sample = response.json()
        response = requests.get(sample["audio_url"])
        return AudioSegment.from_wav(io.BytesIO(response.content))  # type: ignore

    def split_text(self, string):
        # Base case: if the string is less than or equal to MAX_TEXT_LENGTH characters, return it as a single element array
        if len(string) <= MAX_TEXT_LENGTH:
            return [string.strip()]

        # Recursive case: find the index of the last sentence ender in the first MAX_TEXT_LENGTH characters of the string
        sentence_enders = [".", "!", "?"]
        index = -1
        for ender in sentence_enders:
            i = string[:MAX_TEXT_LENGTH].rfind(ender)
            if i > index:
                index = i

        # If there is a sentence ender, split the string at that index plus one and strip any spaces from both parts
        if index != -1:
            first_part = string[: index + 1].strip()
            second_part = string[index + 1 :].strip()

        # If there is no sentence ender, find the index of the last comma in the first MAX_TEXT_LENGTH characters of the string
        else:
            index = string[:MAX_TEXT_LENGTH].rfind(",")
            # If there is a comma, split the string at that index plus one and strip any spaces from both parts
            if index != -1:
                first_part = string[: index + 1].strip()
                second_part = string[index + 1 :].strip()

            # If there is no comma, find the index of the last space in the first MAX_TEXT_LENGTH characters of the string
            else:
                index = string[:MAX_TEXT_LENGTH].rfind(" ")
            # If there is a space, split the string at that index and strip any spaces from both parts
            if index != -1:
                first_part = string[:index].strip()
                second_part = string[index:].strip()

            # If there is no space, split the string at MAX_TEXT_LENGTH characters and strip any spaces from both parts
            else:
                first_part = string[:MAX_TEXT_LENGTH].strip()
                second_part = string[MAX_TEXT_LENGTH:].strip()

        # Append the first part to the result array
        result = [first_part]

        # Call the function recursively on the remaining part of the string and extend the result array with it, unless it is empty
        if second_part != "":
            result.extend(self.split_text(second_part))

        # Return the result array
        return result

    async def async_synthesize(self, text: str) -> AudioSegment:
        # This method is similar to the synthesize method, but it uses async IO to synthesize each chunk in parallel

        # Split the text into chunks of less than MAX_TEXT_LENGTH characters
        text_chunks = self.split_text(text)

        # Create a list of tasks for each chunk using asyncio.create_task()
        tasks = [asyncio.create_task(self.async_synthesize_chunk(chunk)) for chunk in text_chunks]

        # Wait for all tasks to complete using asyncio.gather()
        audio_chunks = await asyncio.gather(*tasks)

        # Concatenate and return the results
        return sum(audio_chunks)  # type: ignore

    async def async_synthesize_chunk(self, text: str) -> AudioSegment:
        url, headers, body = self.get_request(text)

        # Create an aiohttp session and post the request asynchronously using await
        async with aiohttp.ClientSession() as session:
            async with session.post(url, headers=headers, json=body) as response:
                assert response.status == 201, (
                    await response.text() + url + str(headers) + str(body)
                )
                sample = await response.json()
                audio_url = sample["audio_url"]

                # Get the audio data asynchronously using await
                async with session.get(audio_url) as response:
                    assert response.status == 200, "Coqui audio download failed"
                    audio_data = await response.read()

                    # Return an AudioSegment object from the audio data
                    return AudioSegment.from_wav(io.BytesIO(audio_data))  # type: ignore

    def get_request(
        self, text: str
    ) -> typing.Tuple[str, typing.Dict[str, str], typing.Dict[str, object]]:
        url = COQUI_BASE_URL
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        body = {
            "text": text,
            "speed": 1,
        }

        if self.use_xtts:
            url += "/xtts/"
            # If we have a voice prompt, use that instead of the voice ID
            if self.voice_prompt is not None:
                url += "render-from-prompt/"
                body["prompt"] = self.voice_prompt
            else:
                url += "render/"
                body["voice_id"] = self.voice_id
        else:
            if self.voice_prompt is not None:
                url += "/from-prompt/"
                body["prompt"] = self.voice_prompt
            else:
                body["voice_id"] = self.voice_id
        return url, headers, body



import abc
from functools import partial
from typing import List, Optional

from fastapi import APIRouter, Form, Request, Response
from loguru import logger
from pydantic.v1 import BaseModel, Field

from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.agent.default_factory import DefaultAgentFactory
from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.events import RecordingEvent
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import (
    TwilioCallConfig,
    TwilioConfig,
    VonageCallConfig,
    VonageConfig,
)
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.synthesizer.default_factory import DefaultSynthesizerFactory
from vocode.streaming.telephony.client.abstract_telephony_client import AbstractTelephonyClient
from vocode.streaming.telephony.client.twilio_client import TwilioClient
from vocode.streaming.telephony.client.vonage_client import VonageClient
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.telephony.server.router.calls import CallsRouter
from vocode.streaming.telephony.templater import get_connection_twiml
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.transcriber.default_factory import DefaultTranscriberFactory
from vocode.streaming.utils import create_conversation_id
from vocode.streaming.utils.events_manager import EventsManager


class AbstractInboundCallConfig(BaseModel, abc.ABC):
    url: str
    agent_config: AgentConfig
    transcriber_config: Optional[TranscriberConfig] = None
    synthesizer_config: Optional[SynthesizerConfig] = None


class TwilioInboundCallConfig(AbstractInboundCallConfig):
    twilio_config: TwilioConfig


class VonageInboundCallConfig(AbstractInboundCallConfig):
    vonage_config: VonageConfig


class VonageAnswerRequest(BaseModel):
    to: str
    from_: str = Field(..., alias="from")
    uuid: str


class TelephonyServer:
    def __init__(
        self,
        base_url: str,
        config_manager: BaseConfigManager,
        inbound_call_configs: List[AbstractInboundCallConfig] = [],
        transcriber_factory: AbstractTranscriberFactory = DefaultTranscriberFactory(),
        agent_factory: AbstractAgentFactory = DefaultAgentFactory(),
        synthesizer_factory: AbstractSynthesizerFactory = DefaultSynthesizerFactory(),
        events_manager: Optional[EventsManager] = None,
    ):
        self.base_url = base_url
        self.router = APIRouter()
        self.config_manager = config_manager
        self.events_manager = events_manager
        self.router.include_router(
            CallsRouter(
                base_url=base_url,
                config_manager=self.config_manager,
                transcriber_factory=transcriber_factory,
                agent_factory=agent_factory,
                synthesizer_factory=synthesizer_factory,
                events_manager=self.events_manager,
            ).get_router()
        )
        for config in inbound_call_configs:
            self.router.add_api_route(
                config.url,
                self.create_inbound_route(inbound_call_config=config),
                methods=["POST"],
            )
        # vonage requires an events endpoint
        self.router.add_api_route("/events", self.events, methods=["GET", "POST"])
        logger.info(f"Set up events endpoint at https://{self.base_url}/events")

        self.router.add_api_route(
            "/recordings/{conversation_id}", self.recordings, methods=["GET", "POST"]
        )
        logger.info(
            f"Set up recordings endpoint at https://{self.base_url}/recordings/{{conversation_id}}"
        )

    def events(self, request: Request):
        return Response()

    async def recordings(self, request: Request, conversation_id: str):
        recording_url = (await request.json())["recording_url"]
        if self.events_manager is not None and recording_url is not None:
            self.events_manager.publish_event(
                RecordingEvent(recording_url=recording_url, conversation_id=conversation_id)
            )
        return Response()

    def create_inbound_route(
        self,
        inbound_call_config: AbstractInboundCallConfig,
    ):
        async def twilio_route(
            twilio_config: TwilioConfig,
            twilio_sid: str = Form(alias="CallSid"),
            twilio_from: str = Form(alias="From"),
            twilio_to: str = Form(alias="To"),
        ) -> Response:
            call_config = TwilioCallConfig(
                transcriber_config=inbound_call_config.transcriber_config
                or TwilioCallConfig.default_transcriber_config(),
                agent_config=inbound_call_config.agent_config,
                synthesizer_config=inbound_call_config.synthesizer_config
                or TwilioCallConfig.default_synthesizer_config(),
                twilio_config=twilio_config,
                twilio_sid=twilio_sid,
                from_phone=twilio_from,
                to_phone=twilio_to,
                direction="inbound",
            )

            conversation_id = create_conversation_id()
            await self.config_manager.save_config(conversation_id, call_config)
            return get_connection_twiml(base_url=self.base_url, call_id=conversation_id)

        async def vonage_route(vonage_config: VonageConfig, request: Request):
            vonage_answer_request = VonageAnswerRequest.parse_obj(await request.json())
            call_config = VonageCallConfig(
                transcriber_config=inbound_call_config.transcriber_config
                or VonageCallConfig.default_transcriber_config(),
                agent_config=inbound_call_config.agent_config,
                synthesizer_config=inbound_call_config.synthesizer_config
                or VonageCallConfig.default_synthesizer_config(),
                vonage_config=vonage_config,
                vonage_uuid=vonage_answer_request.uuid,
                to_phone=vonage_answer_request.from_,
                from_phone=vonage_answer_request.to,
                direction="inbound",
            )
            conversation_id = create_conversation_id()
            await self.config_manager.save_config(conversation_id, call_config)
            vonage_client = VonageClient(
                base_url=self.base_url,
                maybe_vonage_config=vonage_config,
                record_calls=vonage_config.record,
            )
            return vonage_client.create_call_ncco(
                conversation_id=conversation_id,
                record=vonage_config.record,
            )

        if isinstance(inbound_call_config, TwilioInboundCallConfig):
            logger.info(
                f"Set up inbound call TwiML at https://{self.base_url}{inbound_call_config.url}"
            )
            return partial(twilio_route, inbound_call_config.twilio_config)
        elif isinstance(inbound_call_config, VonageInboundCallConfig):
            logger.info(
                f"Set up inbound call NCCO at https://{self.base_url}{inbound_call_config.url}"
            )
            return partial(vonage_route, inbound_call_config.vonage_config)
        else:
            raise ValueError(f"Unknown inbound call config type {type(inbound_call_config)}")

    async def end_outbound_call(self, conversation_id: str):
        # TODO validation via twilio_client
        call_config = await self.config_manager.get_config(conversation_id)
        if not call_config:
            raise ValueError(f"Could not find call config for {conversation_id}")
        telephony_client: AbstractTelephonyClient
        if isinstance(call_config, TwilioCallConfig):
            telephony_client = TwilioClient(
                base_url=self.base_url, maybe_twilio_config=call_config.twilio_config
            )
            await telephony_client.end_call(call_config.twilio_sid)
        elif isinstance(call_config, VonageCallConfig):
            telephony_client = VonageClient(
                base_url=self.base_url, maybe_vonage_config=call_config.vonage_config
            )
            await telephony_client.end_call(call_config.vonage_uuid)
        return {"id": conversation_id}

    def get_router(self) -> APIRouter:
        return self.router



from typing import Type

from pydantic.v1 import BaseModel

from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.models.actions import ActionConfig as VocodeActionConfig
from vocode.streaming.models.actions import ActionInput, ActionOutput


class WaitVocodeActionConfig(VocodeActionConfig, type="action_wait"):  # type: ignore
    pass


class WaitParameters(BaseModel):
    pass


class WaitResponse(BaseModel):
    success: bool


class Wait(
    BaseAction[
        WaitVocodeActionConfig,
        WaitParameters,
        WaitResponse,
    ]
):
    description: str = (
        "Use this action to wait for the IVR to finish talking or to continue waiting on hold."
    )
    parameters_type: Type[WaitParameters] = WaitParameters
    response_type: Type[WaitResponse] = WaitResponse

    def __init__(
        self,
        action_config: WaitVocodeActionConfig,
    ):
        super().__init__(
            action_config,
            quiet=True,
            should_respond="never",
        )

    async def run(self, action_input: ActionInput[WaitParameters]) -> ActionOutput[WaitResponse]:
        if action_input.user_message_tracker is not None:
            await action_input.user_message_tracker.wait()

        return ActionOutput(
            action_type=action_input.action_config.type,
            response=WaitResponse(success=True),
        )



from enum import Enum
from typing import Optional

from .model import TypedModel


class MessageType(str, Enum):
    BASE = "message_base"
    SSML = "message_ssml"
    BOT_BACKCHANNEL = "bot_backchannel"
    LLM_TOKEN = "llm_token"


class BaseMessage(TypedModel, type=MessageType.BASE):  # type: ignore
    text: str
    trailing_silence_seconds: float = 0.0
    cache_phrase: Optional[str] = None


class SSMLMessage(BaseMessage, type=MessageType.SSML):  # type: ignore
    ssml: str


class BotBackchannel(BaseMessage, type=MessageType.BOT_BACKCHANNEL):  # type: ignore
    pass


class LLMToken(BaseMessage, type=MessageType.LLM_TOKEN):  # type: ignore
    pass


class SilenceMessage(BotBackchannel):
    text: str = "<silence>"
    trailing_silence_seconds: float = 1.0



from enum import Enum


class AudioEncoding(str, Enum):
    LINEAR16 = "linear16"
    MULAW = "mulaw"


class SamplingRate(int, Enum):
    RATE_8000 = 8000
    RATE_16000 = 16000
    RATE_22050 = 22050
    RATE_24000 = 24000
    RATE_44100 = 44100
    RATE_48000 = 48000



import io
from typing import Optional

from elevenlabs import Voice, VoiceSettings
from elevenlabs.client import ElevenLabs
from pydub import AudioSegment

from vocode import getenv
from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer


class ElevenLabsSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        voice_id: str,
        stability: Optional[float] = None,
        similarity_boost: Optional[float] = None,
        api_key: Optional[str] = None,
    ):

        self.voice_id = voice_id
        self.api_key = getenv("ELEVEN_LABS_API_KEY", api_key)

        self.elevenlabs_client = ElevenLabs(
            api_key=self.api_key,
        )
        self.validate_stability_and_similarity_boost(stability, similarity_boost)
        self.stability = stability
        self.similarity_boost = similarity_boost

    def validate_stability_and_similarity_boost(
        self, stability: Optional[float], similarity_boost: Optional[float]
    ) -> None:
        if (stability is None) != (similarity_boost is None):
            raise ValueError("Both stability and similarity_boost must be set or not set.")

    def synthesize(self, text: str) -> AudioSegment:
        if self.stability is not None and self.similarity_boost is not None:
            voice = Voice(
                voice_id=self.voice_id,
                settings=VoiceSettings(
                    stability=self.stability, similarity_boost=self.similarity_boost
                ),
            )
        else:
            voice = Voice(voice_id=self.voice_id)

        audio = self.elevenlabs_client.generate(text=text, voice=voice)

        return AudioSegment.from_mp3(io.BytesIO(audio))  # type: ignore



import io

import boto3
from pydub import AudioSegment

from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer

DEFAULT_SAMPLING_RATE = 16000
DEFAULT_LANGUAGE_CODE = "en-US"
DEFAULT_VOICE_ID = "Matthew"


class PollySynthesizer(BaseSynthesizer):
    def __init__(
        self,
        sampling_rate: int = DEFAULT_SAMPLING_RATE,
        language_code: str = DEFAULT_LANGUAGE_CODE,
        voice_id: str = DEFAULT_VOICE_ID,
    ):
        client = boto3.client("polly")

        # AWS Polly supports sampling rate of 8k and 16k for pcm output
        if sampling_rate not in [8000, 16000]:
            raise Exception(
                "Sampling rate not supported by AWS Polly",
                sampling_rate,
            )

        self.sampling_rate = sampling_rate
        self.client = client
        self.language_code = language_code
        self.voice_id = voice_id

    def synthesize(self, message: str) -> AudioSegment:
        response = self.client.synthesize_speech(
            Text=message,
            LanguageCode=self.language_code,
            TextType="text",
            OutputFormat="pcm",
            VoiceId=self.voice_id,
        )

        return AudioSegment(
            response.get("AudioStream").read(),
            sample_width=2,
            frame_rate=self.sampling_rate,
            channels=1,
        )



import asyncio
import audioop
import base64
from typing import AsyncGenerator, List, Optional, Tuple

import numpy as np
import websockets
from loguru import logger
from pydantic import BaseModel, conint

from vocode.streaming.models.audio import AudioEncoding, SamplingRate
from vocode.streaming.models.message import BaseMessage, BotBackchannel, LLMToken
from vocode.streaming.models.synthesizer import ElevenLabsSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult
from vocode.streaming.synthesizer.eleven_labs_synthesizer import ElevenLabsSynthesizer
from vocode.streaming.synthesizer.input_streaming_synthesizer import InputStreamingSynthesizer

NONCE = "071b5f21-3b24-4427-817e-62508007ae60"
ELEVEN_LABS_BASE_URL = "wss://api.elevenlabs.io/v1/"


# Based on https://github.com/elevenlabs/elevenlabs-python/blob/main/src/elevenlabs/tts.py
async def string_chunker(string: str) -> AsyncGenerator[str, None]:
    splitters = (".", ",", "?", "!", ";", ":", "—", "-", "(", ")", "[", "]", "}", " ")
    buffer = ""
    for text in string:
        if buffer.endswith(splitters):
            yield buffer if buffer.endswith(" ") else buffer + " "
            buffer = text
        elif text.startswith(splitters):
            output = buffer + text[0]
            yield output if output.endswith(" ") else output + " "
            buffer = text[1:]
        else:
            buffer += text
    if buffer != "":
        yield buffer + " "


class ElevenLabsWebsocketVoiceSettings(BaseModel):
    stability: float
    similarity_boost: float


class ElevenLabsWebsocketGenerationConfig(BaseModel):
    chunk_length_schedule: list[conint(ge=50, le=500)]  # type: ignore
    # TODO: Replace above with below for pydantic 2.X
    # chunk_length_schedule: list[Annotated[int, Field(strict=True, ge=50, le=500)]]


class ElevenLabsWebsocketMessage(BaseModel):
    text: str
    try_trigger_generation: bool = True
    flush: bool = False


class ElevenLabsWebsocketFirstMessage(ElevenLabsWebsocketMessage, BaseModel):
    voice_settings: ElevenLabsWebsocketVoiceSettings | None = None
    generation_config: ElevenLabsWebsocketGenerationConfig | None = None
    xi_api_key: str


class ElevenLabsWebsocketResponseAlignment(BaseModel):
    chars: list[str]
    charStartTimesMs: list[int]
    charDurationsMs: list[int]


class ElevenLabsWebsocketResponse(BaseModel):
    audio: Optional[str] = None
    isFinal: Optional[bool] = None
    normalizedAlignment: Optional[ElevenLabsWebsocketResponseAlignment] = None
    alignment: Optional[ElevenLabsWebsocketResponseAlignment] = None

    def __str__(self):
        return f"ElevenLabsWebsocketResponse(has_audio={self.audio is not None}, isFinal={self.isFinal}, alignment={self.alignment})"


class ElevenLabsWSSynthesizer(
    BaseSynthesizer[ElevenLabsSynthesizerConfig], InputStreamingSynthesizer
):
    def __init__(
        self,
        synthesizer_config: ElevenLabsSynthesizerConfig,
    ):
        super().__init__(
            synthesizer_config,
        )

        assert synthesizer_config.api_key is not None, "API key must be set"
        assert synthesizer_config.voice_id is not None, "Voice ID must be set"
        self.api_key = synthesizer_config.api_key
        self.voice_id = synthesizer_config.voice_id
        self.stability = synthesizer_config.stability
        self.similarity_boost = synthesizer_config.similarity_boost
        self.model_id = synthesizer_config.model_id
        self.optimize_streaming_latency = synthesizer_config.optimize_streaming_latency
        self.words_per_minute = 150

        self.text_chunk_queue: asyncio.Queue[Optional[BotBackchannel | LLMToken]] = asyncio.Queue()
        self.voice_packet_queue: asyncio.Queue[Optional[bytes]] = asyncio.Queue()
        self.current_turn_utterances_by_chunk: List[Tuple[str, float]] = []
        self.sample_width = 2 if synthesizer_config.audio_encoding == AudioEncoding.LINEAR16 else 1

        self.websocket_listener: asyncio.Task | None = None
        self.websocket_tasks: dict[str, asyncio.Task | None] = {
            "listener": None,
            "writer": None,
        }
        self.end_of_turn = False

        self.upsample = None
        self.sample_rate = self.synthesizer_config.sampling_rate

        # While this looks useless, we need to assign the response of `asyncio.gather`
        # to *something* or we risk garbage collection of the running coroutines spawned
        # by `asyncio.gather`.
        self.websocket_functions: list[None] | None = None

        if self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            match self.synthesizer_config.sampling_rate:
                case SamplingRate.RATE_16000:
                    self.output_format = "pcm_16000"
                case SamplingRate.RATE_22050:
                    self.output_format = "pcm_22050"
                case SamplingRate.RATE_24000:
                    self.output_format = "pcm_24000"
                case SamplingRate.RATE_44100:
                    self.output_format = "pcm_44100"
                case SamplingRate.RATE_48000:
                    self.output_format = "pcm_44100"
                    self.upsample = SamplingRate.RATE_48000.value
                    self.sample_rate = SamplingRate.RATE_44100.value
                case _:
                    raise ValueError(
                        f"Unsupported sampling rate: {self.synthesizer_config.sampling_rate}. Elevenlabs only supports 16000, 22050, 24000, and 44100 Hz."
                    )
        elif self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            self.output_format = "ulaw_8000"
        else:
            raise ValueError(
                f"Unsupported audio encoding: {self.synthesizer_config.audio_encoding}"
            )

    def get_eleven_labs_websocket_voice_settings(self):
        if self.stability is None or self.similarity_boost is None:
            return None
        return ElevenLabsWebsocketVoiceSettings(
            stability=self.stability,
            similarity_boost=self.similarity_boost,
        )

    def reduce_chunk_amplitude(self, chunk: bytes, factor: float) -> bytes:
        if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            chunk = audioop.ulaw2lin(chunk, 2)
        pcm = np.frombuffer(chunk, dtype=np.int16)
        pcm = (pcm * factor).astype(np.int16)
        pcm_bytes = pcm.tobytes()
        if self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            return audioop.lin2ulaw(pcm_bytes, 2)
        else:
            return pcm_bytes

    async def establish_websocket_listeners(self, chunk_size):
        url = (
            ELEVEN_LABS_BASE_URL
            + f"text-to-speech/{self.voice_id}/stream-input?output_format={self.output_format}"
        )
        if self.optimize_streaming_latency:
            url += f"&optimize_streaming_latency={self.optimize_streaming_latency}"
        if self.model_id:
            url += f"&model_id={self.model_id}"
        headers = {"xi-api-key": self.api_key}

        backchannelled = False

        async with websockets.connect(
            url,
            extra_headers=headers,
        ) as ws:

            async def write() -> None:
                nonlocal backchannelled
                try:
                    first_message = True
                    while True:
                        message = await self.text_chunk_queue.get()
                        if not message:
                            break
                        if first_message and isinstance(message, BotBackchannel):
                            backchannelled = True
                        eleven_labs_ws_message = (
                            ElevenLabsWebsocketMessage(
                                text=message.text,
                                flush=not isinstance(message, LLMToken),
                            ).json()
                            if not first_message
                            else ElevenLabsWebsocketFirstMessage(
                                text=message.text,
                                voice_settings=self.get_eleven_labs_websocket_voice_settings(),
                                generation_config=ElevenLabsWebsocketGenerationConfig(
                                    chunk_length_schedule=[50],
                                ),
                                flush=not isinstance(message, LLMToken),
                                xi_api_key=self.api_key,
                            ).json(exclude_none=True)
                        )
                        await ws.send(eleven_labs_ws_message)
                        first_message = False
                finally:
                    await ws.send(ElevenLabsWebsocketMessage(text="").json())

            async def listen() -> None:
                """Listen to the websocket for audio data and stream it."""

                first_message = True
                buffer = bytearray()
                while True:
                    message = await ws.recv()
                    if "audio" not in message:
                        continue
                    response = ElevenLabsWebsocketResponse.model_validate_json(message)
                    if response.audio:
                        decoded = base64.b64decode(response.audio)
                        seconds = len(decoded) / (
                            self.sample_width * self.synthesizer_config.sampling_rate
                        )

                        if self.upsample:
                            decoded = self._resample_chunk(
                                decoded,
                                self.sample_rate,
                                self.upsample,
                            )
                            seconds = len(decoded) / (self.sample_width * self.sample_rate)

                        if response.alignment:
                            utterance_chunk = "".join(response.alignment.chars) + " "
                            self.current_turn_utterances_by_chunk.append((utterance_chunk, seconds))
                        # For backchannels, send them all as one chunk (so it can't be interrupted) and reduce the volume
                        # so that in the case of a false endpoint, the backchannel is not too loud.
                        if first_message and backchannelled:
                            buffer.extend(decoded)
                            logger.info("First message was a backchannel, reducing volume.")
                            reduced_amplitude_buffer = self.reduce_chunk_amplitude(
                                buffer, factor=self.synthesizer_config.backchannel_amplitude_factor
                            )
                            await self.voice_packet_queue.put(reduced_amplitude_buffer)
                            buffer = bytearray()
                            first_message = False
                        else:
                            buffer.extend(decoded)
                            for chunk_idx in range(0, len(buffer) - chunk_size, chunk_size):
                                await self.voice_packet_queue.put(
                                    buffer[chunk_idx : chunk_idx + chunk_size]
                                )
                            buffer = buffer[len(buffer) - (len(buffer) % chunk_size) :]

                    if response.isFinal:
                        await self.voice_packet_queue.put(None)
                        break

            self.websocket_tasks["listener"] = asyncio.create_task(listen())
            self.websocket_tasks["writer"] = asyncio.create_task(write())
            self.websocket_functions = await asyncio.gather(*self.websocket_tasks.values())

    def get_current_utterance_synthesis_result(self):
        return SynthesisResult(
            self.chunk_result_generator_from_queue(self.voice_packet_queue),
            lambda seconds: self.get_current_message_so_far(seconds),
        )

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ):
        """
        Ran when doing utterance parsing.
        ie: "Hello, my name is foo."
        """
        if not self.websocket_listener:
            self.websocket_listener = asyncio.create_task(
                self.establish_websocket_listeners(chunk_size)
            )

        if isinstance(message, BotBackchannel):
            if not message.text.endswith(" "):
                message.text += " "
            await self.text_chunk_queue.put(message)
            self.total_chars += len(message.text)
        else:
            async for text in string_chunker(message.text):
                await self.text_chunk_queue.put(LLMToken(text=text))
                self.total_chars += len(text)

        return self.get_current_utterance_synthesis_result()

    async def send_token_to_synthesizer(self, message: LLMToken, chunk_size: int):
        """
        Ran when parsing a single chunk of text.

        ie: "Hello,"
        """
        self.total_chars += len(message.text)

        if not self.websocket_listener:
            self.websocket_listener = asyncio.create_task(
                self.establish_websocket_listeners(chunk_size)
            )

        await self.text_chunk_queue.put(message)
        return None

    def _cleanup_websocket_tasks(self):
        for task in self.websocket_tasks.values():
            if task is not None:
                task.cancel()
        self.text_chunk_queue = asyncio.Queue()
        self.voice_packet_queue = asyncio.Queue()
        if self.websocket_listener is not None:
            self.websocket_listener.cancel()

    def ready_synthesizer(self, chunk_size: int):
        self._cleanup_websocket_tasks()
        self.websocket_listener = asyncio.create_task(
            self.establish_websocket_listeners(chunk_size)
        )

    def get_current_message_so_far(self, seconds: float) -> str:
        seconds_idx = 0.0
        buffer = ""
        for utterance, duration in self.current_turn_utterances_by_chunk:
            if seconds_idx > seconds:
                return buffer
            buffer += utterance
            seconds_idx += duration
        return buffer

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: ElevenLabsSynthesizerConfig):
        return ElevenLabsSynthesizer.get_voice_identifier(synthesizer_config)

    async def handle_end_of_turn(self):
        self.end_of_turn = True
        await self.text_chunk_queue.put(None)
        self.current_turn_utterances_by_chunk = []

    async def cancel_websocket_tasks(self):
        self._cleanup_websocket_tasks()

    async def tear_down(self):
        await self.cancel_websocket_tasks()
        await super().tear_down()



from typing import Literal, Type

from loguru import logger
from pydantic.v1 import BaseModel

from vocode.streaming.action.phone_call_action import (
    TwilioPhoneConversationAction,
    VonagePhoneConversationAction,
)
from vocode.streaming.models.actions import ActionConfig as VocodeActionConfig
from vocode.streaming.models.actions import ActionInput, ActionOutput
from vocode.streaming.utils.async_requester import AsyncRequestor
from vocode.streaming.utils.phone_numbers import sanitize_phone_number
from vocode.streaming.utils.state_manager import (
    TwilioPhoneConversationStateManager,
    VonagePhoneConversationStateManager,
)


class TransferCallParameters(BaseModel):
    pass


class TransferCallResponse(BaseModel):
    success: bool


class TransferCallVocodeActionConfig(VocodeActionConfig, type="action_transfer_call"):  # type: ignore
    phone_number: str

    def action_attempt_to_string(self, input: ActionInput) -> str:
        assert isinstance(input.params, TransferCallParameters)
        return f"Attempting to transfer call to {self.phone_number}"

    def action_result_to_string(self, input: ActionInput, output: ActionOutput) -> str:
        assert isinstance(output.response, TransferCallResponse)
        if output.response.success:
            action_description = "Successfully transferred call"
        else:
            action_description = "Did not transfer call because user interrupted"
        return action_description


FUNCTION_DESCRIPTION = "Transfers the call to a new number. This is never used while on hold."
QUIET = True
IS_INTERRUPTIBLE = True
SHOULD_RESPOND: Literal["always"] = "always"


class TwilioTransferCall(
    TwilioPhoneConversationAction[
        TransferCallVocodeActionConfig, TransferCallParameters, TransferCallResponse
    ]
):
    description: str = FUNCTION_DESCRIPTION
    parameters_type: Type[TransferCallParameters] = TransferCallParameters
    response_type: Type[TransferCallResponse] = TransferCallResponse
    conversation_state_manager: TwilioPhoneConversationStateManager

    def __init__(
        self,
        action_config: TransferCallVocodeActionConfig,
    ):
        super().__init__(
            action_config,
            quiet=QUIET,
            is_interruptible=False,
            should_respond=SHOULD_RESPOND,
        )

    async def transfer_call(self, twilio_call_sid: str, to_phone: str):
        twilio_client = self.conversation_state_manager.create_twilio_client()

        url = "https://api.twilio.com/2010-04-01/Accounts/{twilio_account_sid}/Calls/{twilio_call_sid}.json".format(
            twilio_account_sid=twilio_client.get_telephony_config().account_sid,
            twilio_call_sid=twilio_call_sid,
        )

        twiml_data = "<Response><Dial>{to_phone}</Dial></Response>".format(to_phone=to_phone)

        payload = {"Twiml": twiml_data}

        async with AsyncRequestor().get_session() as session:
            async with session.post(url, data=payload, auth=twilio_client.auth) as response:
                if response.status != 200:
                    logger.error(f"Failed to transfer call: {response.status} {response.reason}")
                    raise Exception("failed to update call")
                else:
                    return await response.json()

    async def run(
        self, action_input: ActionInput[TransferCallParameters]
    ) -> ActionOutput[TransferCallResponse]:
        twilio_call_sid = self.get_twilio_sid(action_input)

        sanitized_phone_number = sanitize_phone_number(self.action_config.phone_number)

        if action_input.user_message_tracker is not None:
            await action_input.user_message_tracker.wait()

        logger.info("Finished waiting for user message tracker, now attempting to transfer call")

        if self.conversation_state_manager.transcript.was_last_message_interrupted():
            logger.info("Last bot message was interrupted, not transferring call")
            return ActionOutput(
                action_type=action_input.action_config.type,
                response=TransferCallResponse(success=False),
            )

        await self.transfer_call(twilio_call_sid, sanitized_phone_number)

        return ActionOutput(
            action_type=action_input.action_config.type,
            response=TransferCallResponse(success=True),
        )


class VonageTransferCall(
    VonagePhoneConversationAction[
        TransferCallVocodeActionConfig, TransferCallParameters, TransferCallResponse
    ]
):
    description: str = FUNCTION_DESCRIPTION
    parameters_type: Type[TransferCallParameters] = TransferCallParameters
    response_type: Type[TransferCallResponse] = TransferCallResponse
    conversation_state_manager: VonagePhoneConversationStateManager

    def __init__(self, action_config: TransferCallVocodeActionConfig):
        super().__init__(
            action_config,
            quiet=QUIET,
            is_interruptible=IS_INTERRUPTIBLE,
            should_respond=SHOULD_RESPOND,
        )

    async def run(
        self, action_input: ActionInput[TransferCallParameters]
    ) -> ActionOutput[TransferCallResponse]:
        if action_input.user_message_tracker is not None:
            await action_input.user_message_tracker.wait()
        self.conversation_state_manager.mute_agent()

        sanitized_phone_number = sanitize_phone_number(self.action_config.phone_number)

        if self.conversation_state_manager.get_direction() == "outbound":
            agent_phone_number = self.conversation_state_manager.get_from_phone()
        else:
            agent_phone_number = self.conversation_state_manager.get_to_phone()

        await self.conversation_state_manager.create_vonage_client().update_call(
            vonage_uuid=self.get_vonage_uuid(action_input),
            new_ncco=[
                {
                    "action": "connect",
                    "timeout": "45",
                    "from": agent_phone_number,
                    "endpoint": [
                        {
                            "type": "phone",
                            "number": sanitized_phone_number,
                        }
                    ],
                }
            ],
        )

        return ActionOutput(
            action_type=action_input.action_config.type,
            response=TransferCallResponse(success=True),
        )



import queue

from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import GoogleTranscriberConfig, Transcription
from vocode.streaming.transcriber.base_transcriber import BaseThreadAsyncTranscriber


# TODO: make this nonblocking so it can run in the main thread, see speech.async_client.SpeechAsyncClient
class GoogleTranscriber(BaseThreadAsyncTranscriber[GoogleTranscriberConfig]):
    def __init__(
        self,
        transcriber_config: GoogleTranscriberConfig,
    ):
        super().__init__(transcriber_config)

        import google.auth
        from google.cloud import speech

        google.auth.default()
        self.speech = speech

        self._ended = False
        self.google_streaming_config = self.create_google_streaming_config()
        self.client = self.speech.SpeechClient()
        self.is_ready = False
        if self.transcriber_config.endpointing_config:
            raise Exception("Google endpointing config not supported yet")

    def create_google_streaming_config(self):
        extra_params = {}
        if self.transcriber_config.model:
            extra_params["model"] = self.transcriber_config.model
            extra_params["use_enhanced"] = True

        if self.transcriber_config.language_code:
            extra_params["language_code"] = self.transcriber_config.language_code

        if self.transcriber_config.audio_encoding == AudioEncoding.LINEAR16:
            google_audio_encoding = self.speech.RecognitionConfig.AudioEncoding.LINEAR16
        elif self.transcriber_config.audio_encoding == AudioEncoding.MULAW:
            google_audio_encoding = self.speech.RecognitionConfig.AudioEncoding.MULAW

        return self.speech.StreamingRecognitionConfig(
            config=self.speech.RecognitionConfig(
                encoding=google_audio_encoding,
                sample_rate_hertz=self.transcriber_config.sampling_rate,
                **extra_params
            ),
            interim_results=True,
        )

    def _run_loop(self):
        stream = self.generator()
        requests = (
            self.speech.StreamingRecognizeRequest(audio_content=content) for content in stream
        )
        responses = self.client.streaming_recognize(self.google_streaming_config, requests)
        self.process_responses_loop(responses)

    def terminate(self):
        self._ended = True
        super().terminate()

    def process_responses_loop(self, responses):
        for response in responses:
            self._on_response(response)

            if self._ended:
                break

    def _on_response(self, response):
        if not response.results:
            return

        result = response.results[0]
        if not result.alternatives:
            return

        top_choice = result.alternatives[0]
        message = top_choice.transcript
        confidence = top_choice.confidence

        self.output_janus_queue.sync_q.put_nowait(
            Transcription(message=message, confidence=confidence, is_final=result.is_final)
        )

    def generator(self):
        while not self._ended:
            # Use a blocking get() to ensure there's at least one chunk of
            # data, and stop iteration if the chunk is None, indicating the
            # end of the audio stream.
            chunk = self.input_janus_queue.sync_q.get()
            if chunk is None:
                return
            data = [chunk]

            # Now consume whatever other data's still buffered.
            while True:
                try:
                    chunk = self.input_janus_queue.sync_q.get_nowait()
                    if chunk is None:
                        return
                    data.append(chunk)
                except queue.Empty:
                    break

            yield b"".join(data)



import asyncio
import json
from concurrent.futures import ThreadPoolExecutor
from typing import Any

import boto3

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import PollySynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    SynthesisResult,
    encode_as_wav,
)


class PollySynthesizer(BaseSynthesizer[PollySynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: PollySynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        client = boto3.client("polly")

        # AWS Polly supports sampling rate of 8k and 16k for pcm output
        if synthesizer_config.sampling_rate not in [8000, 16000]:
            raise Exception(
                "Sampling rate not supported by AWS Polly",
                synthesizer_config.sampling_rate,
            )

        self.sampling_rate = synthesizer_config.sampling_rate
        self.client = client
        self.language_code = synthesizer_config.language_code
        self.voice_id = synthesizer_config.voice_id
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    def synthesize(self, message: str) -> Any:
        # Perform the text-to-speech request on the text input with the selected
        # voice parameters and audio file type
        return self.client.synthesize_speech(
            Text=message,
            LanguageCode=self.language_code,
            TextType="text",
            OutputFormat="pcm",
            VoiceId=self.voice_id,
            SampleRate=str(self.sampling_rate),
        )

    def get_speech_marks(self, message: str) -> Any:
        return self.client.synthesize_speech(
            Text=message,
            LanguageCode=self.language_code,
            TextType="text",
            OutputFormat="json",
            VoiceId=self.voice_id,
            SampleRate=str(self.sampling_rate),
            SpeechMarkTypes=["word"],
        )

    # given the number of seconds the message was allowed to go until, where did we get in the message?
    def get_message_up_to(
        self,
        message: str,
        seconds: float,
        word_events,
    ) -> str:
        for event in word_events:
            # time field is in ms
            if event["time"] > seconds * 1000:
                return message[: event["start"]]
        return message

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        audio_response = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor, self.synthesize, message.text
        )
        audio_stream = audio_response.get("AudioStream")

        speech_marks_response = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool_executor, self.get_speech_marks, message.text
        )
        word_events = [
            json.loads(v)
            for v in speech_marks_response.get("AudioStream").read().decode().split()
            if v
        ]

        async def chunk_generator(audio_data_stream, chunk_transform=lambda x: x):
            audio_buffer = await asyncio.get_event_loop().run_in_executor(
                self.thread_pool_executor,
                lambda: audio_stream.read(chunk_size),
            )
            if len(audio_buffer) != chunk_size:
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer), True)
                return
            else:
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer), False)
            while True:
                audio_buffer = audio_stream.read(chunk_size)
                if len(audio_buffer) != chunk_size:
                    yield SynthesisResult.ChunkResult(
                        chunk_transform(audio_buffer[: len(audio_buffer)]), True
                    )
                    break
                yield SynthesisResult.ChunkResult(chunk_transform(audio_buffer), False)

        if self.synthesizer_config.should_encode_as_wav:
            output_generator = chunk_generator(
                audio_stream,
                lambda chunk: encode_as_wav(chunk, self.synthesizer_config),
            )
        else:
            output_generator = chunk_generator(audio_stream)

        return SynthesisResult(
            output_generator,
            lambda seconds: self.get_message_up_to(
                message.text,
                seconds,
                word_events,
            ),
        )



import aiohttp
import httpx

from vocode.streaming.utils.singleton import Singleton


class AsyncRequestor(Singleton):
    def __init__(self):
        self.session = aiohttp.ClientSession()
        self.async_client = httpx.AsyncClient()

    def get_session(self):
        if self.session.closed:
            self.session = aiohttp.ClientSession()
        return self.session

    def get_client(self):
        return self.async_client

    async def close_session(self):
        if not self.session.closed:
            await self.session.close()

    async def close_client(self):
        await self.async_client.aclose()



import io
import wave
import hashlib

from vocode import getenv
from vocode.streaming.models.audio import AudioEncoding, SamplingRate
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import CartesiaSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log


class CartesiaSynthesizer(BaseSynthesizer[CartesiaSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: CartesiaSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        # Lazy import the cartesia module
        try:
            from cartesia.tts import AsyncCartesiaTTS
        except ImportError as e:
            raise ImportError(
                f"Missing required dependancies for CartesiaSynthesizer"
            ) from e
        
        self.cartesia_tts = AsyncCartesiaTTS
        
        self.api_key = synthesizer_config.api_key or getenv("CARTESIA_API_KEY")
        if not self.api_key:
            raise ValueError("Missing Cartesia API key")
        

        if synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            self.channel_width = 2
            match synthesizer_config.sampling_rate:
                case SamplingRate.RATE_44100:
                    self.sampling_rate = 44100
                    self.output_format = "pcm_44100"
                case SamplingRate.RATE_22050:
                    self.sampling_rate = 22050
                    self.output_format = "pcm_22050"
                case SamplingRate.RATE_16000:
                    self.sampling_rate = 16000
                    self.output_format = "pcm_16000"
                case _:
                    raise ValueError(
                        f"Unsupported PCM sampling rate {synthesizer_config.sampling_rate}"
                    )
        elif synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            # Cartesia has issues with MuLaw/8000. Use pcm/16000 and
            # create_synthesis_result_from_wav will handle the conversion to mulaw/8000
            self.channel_width = 2
            self.output_format = "pcm_16000"
            self.sampling_rate = 16000
        else:
            raise ValueError(
                f"Unsupported audio encoding {synthesizer_config.audio_encoding}"
            )

        self.num_channels = 1
        self.model_id = synthesizer_config.model_id
        self.voice_id = synthesizer_config.voice_id
        self.client = self.cartesia_tts(api_key=self.api_key)
        self.voice_embedding = self.client.get_voice_embedding(voice_id=self.voice_id)
        

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        generator = await self.client.generate(
            transcript=message.text,
            voice=self.voice_embedding,
            stream=True,
            model_id=self.model_id,
            data_rtype='bytes',
            output_format=self.output_format
        )

        audio_file = io.BytesIO()
        with wave.open(audio_file, 'wb') as wav_file:
            wav_file.setnchannels(self.num_channels)
            wav_file.setsampwidth(self.channel_width)
            wav_file.setframerate(self.sampling_rate)
            async for chunk in generator:
                wav_file.writeframes(chunk['audio'])
        audio_file.seek(0)

        result = self.create_synthesis_result_from_wav(
            synthesizer_config=self.synthesizer_config,
            file=audio_file,
            message=message,
            chunk_size=chunk_size,
        )

        return result
    
    @classmethod
    def get_voice_identifier(cls, synthesizer_config: CartesiaSynthesizerConfig):
        hashed_api_key = hashlib.sha256(f"{synthesizer_config.api_key}".encode("utf-8")).hexdigest()
        return ":".join(
            (
                "cartesia",
                hashed_api_key,
                str(synthesizer_config.voice_id),
                str(synthesizer_config.model_id),
                synthesizer_config.audio_encoding
            )
        )


import asyncio
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO

from gtts import gTTS
from pydub import AudioSegment

from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import GTTSSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult


class GTTSSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        synthesizer_config: GTTSSynthesizerConfig,
    ):
        super().__init__(synthesizer_config)

        self.thread_pool_executor = ThreadPoolExecutor(max_workers=1)

    async def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        audio_file = BytesIO()

        def thread():
            tts = gTTS(message.text)
            tts.write_to_fp(audio_file)

        await asyncio.get_event_loop().run_in_executor(self.thread_pool_executor, thread)
        audio_file.seek(0)
        # TODO: probably needs to be in a thread
        audio_segment: AudioSegment = AudioSegment.from_mp3(audio_file)  # type: ignore
        output_bytes_io = BytesIO()
        audio_segment.export(output_bytes_io, format="wav")  # type: ignore

        result = self.create_synthesis_result_from_wav(
            synthesizer_config=self.synthesizer_config,
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )
        return result



import base64
from enum import Enum
from typing import Optional

from .agent import AgentConfig
from .events import Sender
from .model import TypedModel
from .synthesizer import SynthesizerConfig
from .transcriber import TranscriberConfig
from .transcript import TranscriptEvent
from vocode.streaming.models.client_backend import InputAudioConfig, OutputAudioConfig


class WebSocketMessageType(str, Enum):
    BASE = "websocket_base"
    START = "websocket_start"
    AUDIO = "websocket_audio"
    TRANSCRIPT = "websocket_transcript"
    READY = "websocket_ready"
    STOP = "websocket_stop"
    AUDIO_CONFIG_START = "websocket_audio_config_start"


class WebSocketMessage(TypedModel, type=WebSocketMessageType.BASE):  # type: ignore
    pass


class AudioMessage(WebSocketMessage, type=WebSocketMessageType.AUDIO):  # type: ignore
    data: str

    @classmethod
    def from_bytes(cls, chunk: bytes):
        return cls(data=base64.b64encode(chunk).decode("utf-8"))

    def get_bytes(self) -> bytes:
        return base64.b64decode(self.data)


class TranscriptMessage(WebSocketMessage, type=WebSocketMessageType.TRANSCRIPT):  # type: ignore
    text: str
    sender: Sender
    timestamp: float

    @classmethod
    def from_event(cls, event: TranscriptEvent):
        return cls(text=event.text, sender=event.sender, timestamp=event.timestamp)


class StartMessage(WebSocketMessage, type=WebSocketMessageType.START):  # type: ignore
    transcriber_config: TranscriberConfig
    agent_config: AgentConfig
    synthesizer_config: SynthesizerConfig
    conversation_id: Optional[str] = None


class AudioConfigStartMessage(
    WebSocketMessage, type=WebSocketMessageType.AUDIO_CONFIG_START  # type: ignore
):
    input_audio_config: InputAudioConfig
    output_audio_config: OutputAudioConfig
    conversation_id: Optional[str] = None
    subscribe_transcript: Optional[bool] = None


class ReadyMessage(WebSocketMessage, type=WebSocketMessageType.READY):  # type: ignore
    pass


class StopMessage(WebSocketMessage, type=WebSocketMessageType.STOP):  # type: ignore
    pass



from typing import Optional

from loguru import logger
from redis import Redis

from vocode.streaming.models.telephony import BaseCallConfig
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.utils.redis import initialize_redis


class RedisConfigManager(BaseConfigManager):
    def __init__(self):
        self.redis: Redis = initialize_redis()

    async def _set_with_one_day_expiration(self, *args, **kwargs):
        ONE_DAY_SECONDS = 60 * 60 * 24
        return await self.redis.set(*args, **{**kwargs, "ex": ONE_DAY_SECONDS})

    async def save_config(self, conversation_id: str, config: BaseCallConfig):
        logger.debug(f"Saving config for {conversation_id}")
        await self._set_with_one_day_expiration(conversation_id, config.json())

    async def get_config(self, conversation_id) -> Optional[BaseCallConfig]:
        logger.debug(f"Getting config for {conversation_id}")
        raw_config = await self.redis.get(conversation_id)  # type: ignore
        if raw_config:
            return BaseCallConfig.parse_raw(raw_config)
        return None

    async def delete_config(self, conversation_id):
        logger.debug(f"Deleting config for {conversation_id}")
        await self.redis.delete(conversation_id)



import asyncio
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Literal, Optional, Type, TypeVar

import jsonschema
from jsonschema import Draft202012Validator as SchemaValidator

from vocode.streaming.action.action_utils import exclude_keys_recursive
from vocode.streaming.models.actions import (
    ActionConfig,
    ActionInput,
    ActionOutput,
    ParametersType,
    ResponseType,
)

if TYPE_CHECKING:
    from vocode.streaming.utils.state_manager import AbstractConversationStateManager

ActionConfigType = TypeVar("ActionConfigType", bound=ActionConfig)


class BaseAction(Generic[ActionConfigType, ParametersType, ResponseType]):  # type: ignore
    description: str = ""

    def __init__(
        self,
        action_config: ActionConfigType,
        should_respond: Literal["always", "sometimes", "never"] = "never",
        quiet: bool = False,
        is_interruptible: bool = True,
        **kwargs,
    ):
        self.action_config = action_config
        self.should_respond = should_respond
        self.quiet = quiet
        self.is_interruptible = is_interruptible

    def attach_conversation_state_manager(
        self, conversation_state_manager: "AbstractConversationStateManager"
    ):
        self.conversation_state_manager = conversation_state_manager

    async def run(self, action_input: ActionInput[ParametersType]) -> ActionOutput[ResponseType]:
        raise NotImplementedError

    @property
    def parameters_type(self) -> Type[ParametersType]:
        raise NotImplementedError

    @property
    def response_type(self) -> Type[ResponseType]:
        raise NotImplementedError

    def get_parameters_schema(self) -> Dict[str, Any]:
        return self.parameters_type.schema()

    def get_function_name(self) -> str:
        return self.action_config.type

    def get_openai_function(self):
        parameters_schema = self.get_parameters_schema()
        parameters_schema = exclude_keys_recursive(parameters_schema, {"title"})
        if self.should_respond != "never":
            parameters_schema["properties"]["user_message"] = self._user_message_param_info()
            if self.should_respond == "always":
                required: List[str] = parameters_schema.get("required", [])
                required.append("user_message")
                parameters_schema["required"] = required
        try:
            SchemaValidator.check_schema(parameters_schema)
        except jsonschema.exceptions.SchemaError as e:
            raise ValueError(f"Invalid parameters schema for {self.action_config.type}: {e}") from e

        return {
            "name": self.get_function_name(),
            "description": self.description,
            "parameters": parameters_schema,
        }

    def create_action_params(self, params: Dict[str, Any]) -> ParametersType:
        return self.parameters_type(**params)

    def create_action_input(
        self,
        conversation_id: str,
        params: Dict[str, Any],
        user_message_tracker: Optional[asyncio.Event] = None,
    ) -> ActionInput[ParametersType]:
        if "user_message" in params:
            del params["user_message"]
        return ActionInput(
            action_config=self.action_config,
            conversation_id=conversation_id,
            params=self.create_action_params(params),
            user_message_tracker=user_message_tracker,
        )

    def _user_message_param_info(self):
        return {
            "type": "string",
            "description": """A message to reply to the user with BEFORE we make the function call.
                    Essentially a live response informing them that the function is about to happen.
                    Eg Let me check the weather in San Francisco CA for you """,
        }



from typing import Optional

from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import TwilioConfig
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.telephony.conversation.outbound_call import OutboundCall


class ZoomDialIn(OutboundCall):
    def __init__(
        self,
        base_url: str,
        zoom_number: str,
        zoom_meeting_id: str,
        zoom_meeting_password: Optional[str],
        from_phone: str,
        config_manager: BaseConfigManager,
        agent_config: AgentConfig,
        twilio_config: TwilioConfig,
        transcriber_config: Optional[TranscriberConfig] = None,
        synthesizer_config: Optional[SynthesizerConfig] = None,
        conversation_id: Optional[str] = None,
    ):
        digits = f"wwww{zoom_meeting_id}#"
        if zoom_meeting_password:
            digits += f"wwwwwwww*{zoom_meeting_password}#"

        super().__init__(
            base_url=base_url,
            to_phone=zoom_number,
            from_phone=from_phone,
            config_manager=config_manager,
            transcriber_config=transcriber_config,
            agent_config=agent_config,
            synthesizer_config=synthesizer_config,
            telephony_config=twilio_config,
            conversation_id=conversation_id,
            digits=digits,
        )

        self.zoom_number = zoom_number
        self.zoom_meeting_id = zoom_meeting_id
        self.zoom_meeting_password = zoom_meeting_password



from __future__ import annotations

import asyncio
from typing import List

from loguru import logger

from vocode.streaming.models.events import Event, EventType


class EventsManager:
    def __init__(self, subscriptions: List[EventType] = []):
        self.queue: asyncio.Queue[Event] = asyncio.Queue()
        self.subscriptions = set(subscriptions)
        self.active = False

    def publish_event(self, event: Event):
        if event and event.type in self.subscriptions:
            self.queue.put_nowait(event)

    async def start(self):
        self.active = True
        while self.active:
            try:
                event = await self.queue.get()
            except asyncio.QueueEmpty:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
            await self.handle_event(event)

    async def handle_event(self, event: Event):
        pass

    async def flush(self):
        self.active = False
        while True:
            try:
                event = self.queue.get_nowait()
                await self.handle_event(event)
            except asyncio.QueueEmpty:
                break
            except TypeError as e:
                if "NoneType can't be used in 'await' expression" in str(e):
                    logger.error(
                        "Handle event was overridden with non-async function. Please override with async function."
                    )
                else:
                    raise e from e



from __future__ import annotations

import asyncio
import audioop
from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, Union

from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.transcriber import TranscriberConfig, Transcription
from vocode.streaming.utils.speed_manager import SpeedManager
from vocode.streaming.utils.worker import AsyncWorker, ThreadAsyncWorker

TranscriberConfigType = TypeVar("TranscriberConfigType", bound=TranscriberConfig)


class AbstractTranscriber(Generic[TranscriberConfigType], ABC):
    def __init__(self, transcriber_config: TranscriberConfigType):
        self.transcriber_config = transcriber_config
        self.is_muted = False
        self.speed_manager: Optional[SpeedManager] = None
        self.input_queue: asyncio.Queue[bytes] = asyncio.Queue()
        self.output_queue: asyncio.Queue[Transcription] = asyncio.Queue()

    def attach_speed_manager(self, speed_manager: SpeedManager):
        self.speed_manager = speed_manager

    def mute(self):
        self.is_muted = True

    def unmute(self):
        self.is_muted = False

    def get_transcriber_config(self) -> TranscriberConfigType:
        return self.transcriber_config

    async def ready(self):
        return True

    def create_silent_chunk(self, chunk_size, sample_width=2):
        linear_audio = b"\0" * chunk_size
        if self.get_transcriber_config().audio_encoding == AudioEncoding.LINEAR16:
            return linear_audio
        elif self.get_transcriber_config().audio_encoding == AudioEncoding.MULAW:
            return audioop.lin2ulaw(linear_audio, sample_width)

    @abstractmethod
    async def _run_loop(self):
        pass

    def send_audio(self, chunk):
        if not self.is_muted:
            self.consume_nonblocking(chunk)
        else:
            self.consume_nonblocking(self.create_silent_chunk(len(chunk)))

    @abstractmethod
    def terminate(self):
        pass


class BaseAsyncTranscriber(AbstractTranscriber[TranscriberConfigType], AsyncWorker):
    def __init__(self, transcriber_config: TranscriberConfigType):
        AbstractTranscriber.__init__(self, transcriber_config)
        AsyncWorker.__init__(self, self.input_queue, self.output_queue)

    async def _run_loop(self):
        raise NotImplementedError

    def terminate(self):
        AsyncWorker.terminate(self)


class BaseThreadAsyncTranscriber(AbstractTranscriber[TranscriberConfigType], ThreadAsyncWorker):
    def __init__(self, transcriber_config: TranscriberConfigType):
        AbstractTranscriber.__init__(self, transcriber_config)
        ThreadAsyncWorker.__init__(self, self.input_queue, self.output_queue)

    def _run_loop(self):
        raise NotImplementedError

    def terminate(self):
        ThreadAsyncWorker.terminate(self)


BaseTranscriber = Union[
    BaseAsyncTranscriber[TranscriberConfigType],
    BaseThreadAsyncTranscriber[TranscriberConfigType],
]



import asyncio
from typing import Optional

from fastapi import WebSocket

from vocode.streaming.output_device.base_output_device import BaseOutputDevice
from vocode.streaming.output_device.speaker_output import SpeakerOutput
from vocode.streaming.telephony.constants import (
    PCM_SILENCE_BYTE,
    VONAGE_AUDIO_ENCODING,
    VONAGE_CHUNK_SIZE,
    VONAGE_SAMPLING_RATE,
)
from vocode.streaming.utils.create_task import asyncio_create_task_with_done_error_log


class VonageOutputDevice(BaseOutputDevice):
    def __init__(
        self,
        ws: Optional[WebSocket] = None,
        output_to_speaker: bool = False,
    ):
        super().__init__(sampling_rate=VONAGE_SAMPLING_RATE, audio_encoding=VONAGE_AUDIO_ENCODING)
        self.ws = ws
        self.active = True
        self.queue: asyncio.Queue[bytes] = asyncio.Queue()
        self.process_task = asyncio_create_task_with_done_error_log(self.process())
        self.output_to_speaker = output_to_speaker
        if output_to_speaker:
            self.output_speaker = SpeakerOutput.from_default_device(
                sampling_rate=VONAGE_SAMPLING_RATE, blocksize=VONAGE_CHUNK_SIZE // 2
            )

    async def process(self):
        while self.active:
            chunk = await self.queue.get()
            if self.output_to_speaker:
                self.output_speaker.consume_nonblocking(chunk)
            for i in range(0, len(chunk), VONAGE_CHUNK_SIZE):
                subchunk = chunk[i : i + VONAGE_CHUNK_SIZE]
                if len(subchunk) % 2 == 1:
                    subchunk += PCM_SILENCE_BYTE  # pad with silence, Vonage goes crazy otherwise
                await self.ws.send_bytes(subchunk)

    def consume_nonblocking(self, chunk: bytes):
        self.queue.put_nowait(chunk)

    def terminate(self):
        self.process_task.cancel()



from typing import TYPE_CHECKING, Optional

import aiohttp

from vocode.streaming.models.vector_db import PineconeConfig, VectorDBConfig
from vocode.streaming.vector_db.base_vector_db import VectorDB

if TYPE_CHECKING:
    from vocode.streaming.vector_db.pinecone import PineconeDB


class VectorDBFactory:
    def create_vector_db(
        self,
        vector_db_config: VectorDBConfig,
        aiohttp_session: Optional[aiohttp.ClientSession] = None,
    ) -> VectorDB:
        if isinstance(vector_db_config, PineconeConfig):
            return self._get_pinecone_db(vector_db_config, aiohttp_session)
        raise Exception("Invalid vector db config", vector_db_config.type)

    def _get_pinecone_db(
        self, vector_db_config: PineconeConfig, aiohttp_session: Optional[aiohttp.ClientSession]
    ) -> "PineconeDB":
        try:
            from vocode.streaming.vector_db.pinecone import PineconeDB

            return PineconeDB(vector_db_config, aiohttp_session=aiohttp_session)
        except ImportError as e:
            raise ImportError(
                f"Missing required dependancies for VectorDB {vector_db_config.type}"
            ) from e



from __future__ import annotations

import asyncio
import queue
from typing import Tuple, Union

import miniaudio
from loguru import logger

from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.utils import convert_wav
from vocode.streaming.utils.mp3_helper import decode_mp3
from vocode.streaming.utils.worker import ThreadAsyncWorker


class MiniaudioWorker(ThreadAsyncWorker[Union[bytes, None]]):
    def __init__(
        self,
        synthesizer_config: SynthesizerConfig,
        chunk_size: int,
        input_queue: asyncio.Queue[Union[bytes, None]],
        output_queue: asyncio.Queue[Tuple[bytes, bool]],
    ) -> None:
        super().__init__(input_queue, output_queue)
        self.output_queue = output_queue  # for typing
        self.synthesizer_config = synthesizer_config
        self.chunk_size = chunk_size
        self._ended = False

    def _run_loop(self):
        # tracks the mp3 so far
        current_mp3_buffer = bytearray()
        # tracks the wav so far
        current_wav_buffer = bytearray()
        # the leftover chunks of the wav that haven't been sent to the output queue yet
        current_wav_output_buffer = bytearray()
        while not self._ended:
            # Get a tuple of (mp3_chunk, is_last) from the input queue
            try:
                mp3_chunk = self.input_janus_queue.sync_q.get(timeout=1)
            except queue.Empty:
                continue
            if mp3_chunk is None:
                current_mp3_buffer.clear()
                current_wav_buffer.clear()
                self.output_janus_queue.sync_q.put((bytes(current_wav_output_buffer), True))
                current_wav_output_buffer.clear()
                continue
            try:
                current_mp3_buffer.extend(mp3_chunk)
                output_bytes = decode_mp3(bytes(current_mp3_buffer))
            except miniaudio.DecodeError as e:
                # TODO: better logging
                logger.exception("MiniaudioWorker error: " + str(e), exc_info=True)
                self.output_janus_queue.sync_q.put(
                    (bytes(current_wav_output_buffer), True)
                )  # sentinel
                continue
            converted_output_bytes = convert_wav(
                output_bytes,
                output_sample_rate=self.synthesizer_config.sampling_rate,
                output_encoding=self.synthesizer_config.audio_encoding,
            )
            # take the difference between the current_wav_buffer and the converted_output_bytes
            # and put the difference in the output buffer
            new_bytes = converted_output_bytes[len(current_wav_buffer) :]
            current_wav_output_buffer.extend(new_bytes)

            # chunk up new_bytes in chunks of chunk_size bytes, but keep the last chunk (less than chunk size) in the wav output buffer
            output_buffer_idx = 0
            while output_buffer_idx < len(current_wav_output_buffer) - self.chunk_size:
                chunk = current_wav_output_buffer[
                    output_buffer_idx : output_buffer_idx + self.chunk_size
                ]
                self.output_janus_queue.sync_q.put(
                    (chunk, False)
                )  # don't need to use bytes() since we already sliced it (which is a copy)
                output_buffer_idx += self.chunk_size

            current_wav_output_buffer = current_wav_output_buffer[output_buffer_idx:]
            current_wav_buffer.extend(new_bytes)

    def terminate(self):
        self._ended = True
        super().terminate()



from typing import Dict, Optional

from loguru import logger

from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import (
    TelephonyConfig,
    TwilioCallConfig,
    TwilioConfig,
    VonageCallConfig,
    VonageConfig,
)
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.telephony.client.abstract_telephony_client import AbstractTelephonyClient
from vocode.streaming.telephony.client.twilio_client import TwilioClient
from vocode.streaming.telephony.client.vonage_client import VonageClient
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.utils import create_conversation_id


class OutboundCall:
    def __init__(
        self,
        base_url: str,
        to_phone: str,
        from_phone: str,
        config_manager: BaseConfigManager,
        agent_config: AgentConfig,
        telephony_config: TelephonyConfig,
        telephony_params: Optional[Dict[str, str]] = None,
        transcriber_config: Optional[TranscriberConfig] = None,
        synthesizer_config: Optional[SynthesizerConfig] = None,
        conversation_id: Optional[str] = None,
        sentry_tags: Dict[str, str] = {},
        digits: Optional[
            str
        ] = None,  # Keys to press when the call connects, see send_digits https://www.twilio.com/docs/voice/api/call-resource#create-a-call-resource
        output_to_speaker: bool = False,
    ):
        self.base_url = base_url
        self.to_phone = to_phone
        self.from_phone = from_phone
        self.config_manager = config_manager
        self.agent_config = agent_config
        self.conversation_id = conversation_id or create_conversation_id()
        self.telephony_config = telephony_config
        self.telephony_params = telephony_params or {}
        self.telephony_client = self.create_telephony_client()
        self.transcriber_config = self.create_transcriber_config(transcriber_config)
        self.synthesizer_config = self.create_synthesizer_config(synthesizer_config)
        self.output_to_speaker = output_to_speaker
        self.sentry_tags = sentry_tags
        self.digits = digits

    def create_telephony_client(self) -> AbstractTelephonyClient:
        if isinstance(self.telephony_config, TwilioConfig):
            return TwilioClient(base_url=self.base_url, maybe_twilio_config=self.telephony_config)
        elif isinstance(self.telephony_config, VonageConfig):
            return VonageClient(base_url=self.base_url, maybe_vonage_config=self.telephony_config)

    def create_transcriber_config(
        self, transcriber_config_override: Optional[TranscriberConfig]
    ) -> TranscriberConfig:
        if transcriber_config_override is not None:
            return transcriber_config_override
        if isinstance(self.telephony_config, TwilioConfig):
            return TwilioCallConfig.default_transcriber_config()
        elif isinstance(self.telephony_config, VonageConfig):
            return VonageCallConfig.default_transcriber_config()
        else:
            raise ValueError("No telephony config provided")

    def create_synthesizer_config(
        self, synthesizer_config_override: Optional[SynthesizerConfig]
    ) -> SynthesizerConfig:
        if synthesizer_config_override is not None:
            return synthesizer_config_override
        if isinstance(self.telephony_config, TwilioConfig):
            return TwilioCallConfig.default_synthesizer_config()
        elif isinstance(self.telephony_config, VonageConfig):
            return VonageCallConfig.default_synthesizer_config()
        else:
            raise ValueError("No telephony config provided")

    async def start(self):
        logger.debug("Starting outbound call")
        self.telephony_id = await self.telephony_client.create_call(
            conversation_id=self.conversation_id,
            to_phone=self.to_phone,
            from_phone=self.from_phone,
            record=self.telephony_client.get_telephony_config().record,  # note twilio does not use this
            telephony_params=self.telephony_params,
            digits=self.digits,
        )
        if isinstance(self.telephony_client, TwilioClient):
            call_config = TwilioCallConfig(
                transcriber_config=self.transcriber_config,
                agent_config=self.agent_config,
                synthesizer_config=self.synthesizer_config,
                twilio_config=self.telephony_client.twilio_config,
                twilio_sid=self.telephony_id,
                from_phone=self.from_phone,
                to_phone=self.to_phone,
                sentry_tags=self.sentry_tags,
                telephony_params=self.telephony_params,
                direction="outbound",
            )
        elif isinstance(self.telephony_client, VonageClient):
            call_config = VonageCallConfig(
                transcriber_config=self.transcriber_config,
                agent_config=self.agent_config,
                synthesizer_config=self.synthesizer_config,
                vonage_config=self.telephony_client.vonage_config,
                vonage_uuid=self.telephony_id,
                from_phone=self.from_phone,
                to_phone=self.to_phone,
                output_to_speaker=False,
                sentry_tags=self.sentry_tags,
                telephony_params=self.telephony_params,
                direction="outbound",
            )
        else:
            raise ValueError("Unknown telephony client")
        await self.config_manager.save_config(self.conversation_id, call_config)

    async def end(self):
        return await self.telephony_client.end_call(self.telephony_id)



import re
from typing import AsyncGenerator, AsyncIterable, List, Literal, Optional, Union

from sentry_sdk.tracing import Span

from vocode.streaming.models.actions import FunctionCall, FunctionFragment

TOKENS_TO_GENERATE_PAST_PERIOD = 3
SENTENCE_ENDINGS_EXCEPT_PERIOD_PATTERN = r"[?!\n\t\r]"


SHORT_SENTENCE_CUTOFF = 3


def split_sentences(text: str) -> List[str]:
    """Splits text into sentences and preserve trailing periods.

    Merge sentences that are just numbers, as they are part of lists.
    """
    initial_split = text.split(". ")

    final_split = []
    buffer = ""

    for i, sentence in enumerate(initial_split):
        is_last = i == len(initial_split) - 1
        buffer += sentence
        if not is_last:
            buffer += ". "
        if not re.fullmatch(r"\d+", sentence.strip()):
            final_split.append(buffer.strip())
            buffer = ""

    if buffer.strip():
        final_split.append(buffer.strip())

    return [sentence for sentence in final_split if sentence]


async def collate_response_async(
    conversation_id: str,
    gen: AsyncIterable[Union[str, FunctionFragment]],
    get_functions: Literal[True, False] = False,
    sentry_span: Optional[Span] = None,
) -> AsyncGenerator[
    Union[str, FunctionCall],
    None,
]:  # tuple of message to send and whether it's the final message
    buffer = ""
    function_name_buffer = ""
    function_args_buffer = ""
    is_post_period = False
    tokens_since_period = 0
    is_first = True
    async for token in gen:
        if is_first:
            if sentry_span:
                sentry_span.finish()
            is_first = False
        if not token:
            continue
        if isinstance(token, str):
            buffer += token
            if len(buffer.strip().split()) < SHORT_SENTENCE_CUTOFF:
                continue
            if re.search(SENTENCE_ENDINGS_EXCEPT_PERIOD_PATTERN, token):
                # split on last occurrence of sentence ending
                matches = [
                    match for match in re.finditer(SENTENCE_ENDINGS_EXCEPT_PERIOD_PATTERN, buffer)
                ]
                last_match = matches[-1]
                split_point = last_match.start() + 1
                to_keep, to_return = buffer[split_point:], buffer[:split_point]
                if to_return.strip():
                    yield to_return.strip()
                buffer = to_keep
            elif "." in token:
                is_post_period = True
                tokens_since_period = 0

            if is_post_period and tokens_since_period > TOKENS_TO_GENERATE_PAST_PERIOD:
                sentences = split_sentences(buffer)
                if len(sentences) > 1:
                    yield " ".join(sentences[:-1])
                    buffer = sentences[-1]
                is_post_period = False
                tokens_since_period = 0
            else:
                tokens_since_period += 1

        elif isinstance(token, FunctionFragment):
            function_name_buffer += token.name
            function_args_buffer += token.arguments
    to_return = buffer.strip()
    if to_return:
        yield to_return
    if function_name_buffer and get_functions:
        yield FunctionCall(name=function_name_buffer, arguments=function_args_buffer)


async def stream_response_async(
    conversation_id: str,
    gen: AsyncIterable[Union[str, FunctionFragment]],
    get_functions: Literal[True, False] = False,
    sentry_span: Optional[Span] = None,
) -> AsyncGenerator[
    Union[str, FunctionCall],
    None,
]:  # tuple of message to send and whether it's the final message
    splitters = (".", ",", "?", "!", ";", ":", "—", "-", "(", ")", "[", "]", "}", " ")
    buffer = ""
    function_name_buffer = ""
    function_args_buffer = ""
    is_first = True
    async for token in gen:
        if is_first:
            if sentry_span:
                sentry_span.finish()
            is_first = False
        if not token:
            continue
        if isinstance(token, str):
            if buffer.endswith(splitters):
                yield buffer if buffer.endswith(" ") else buffer + " "
                buffer = token
            elif token.startswith(splitters):
                output = buffer + token[0]
                yield output if output.endswith(" ") else output + " "
                buffer = token[1:]
            else:
                buffer += token

        elif isinstance(token, FunctionFragment):
            function_name_buffer += token.name
            function_args_buffer += token.arguments
    if buffer != "":
        yield buffer + " "
    if function_name_buffer and get_functions:
        yield FunctionCall(name=function_name_buffer, arguments=function_args_buffer)



from vocode.streaming.agent.openai_utils import merge_event_logs
from vocode.streaming.models.transcript import ActionStart, EventLog, Transcript


def format_anthropic_chat_messages_from_transcript(
    transcript: Transcript,
) -> list[dict]:
    # merge consecutive bot messages
    new_event_logs: list[EventLog] = merge_event_logs(event_logs=transcript.event_logs)

    # Removing BOT_ACTION_START so that it doesn't confuse the completion-y prompt, e.g.
    # BOT: BOT_ACTION_START: action_end_conversation
    # Right now, this version of context does not work for normal actions, only phrase trigger actions

    merged_event_logs_sans_bot_action_start = [
        event_log for event_log in new_event_logs if not isinstance(event_log, ActionStart)
    ]

    return [
        {
            "role": "user",
            "content": Transcript(event_logs=merged_event_logs_sans_bot_action_start).to_string(
                include_timestamps=False,
                mark_human_backchannels_with_brackets=True,
            ),
        },
        {"role": "assistant", "content": "BOT:"},
    ]
    # TODO: reliably count tokens of Anthropic messages so that we don't exceed the context window



from enum import Enum
from typing import Any, Dict, List, Literal, Optional

from pydantic.v1 import validator

from .audio import AudioEncoding, SamplingRate
from .model import BaseModel, TypedModel
from vocode.streaming.models.client_backend import OutputAudioConfig
from vocode.streaming.output_device.base_output_device import BaseOutputDevice
from vocode.streaming.telephony.constants import DEFAULT_AUDIO_ENCODING, DEFAULT_SAMPLING_RATE


class SynthesizerType(str, Enum):
    BASE = "synthesizer_base"
    AZURE = "synthesizer_azure"
    GOOGLE = "synthesizer_google"
    ELEVEN_LABS = "synthesizer_eleven_labs"
    RIME = "synthesizer_rime"
    PLAY_HT = "synthesizer_play_ht"
    GTTS = "synthesizer_gtts"
    STREAM_ELEMENTS = "synthesizer_stream_elements"
    COQUI_TTS = "synthesizer_coqui_tts"
    COQUI = "synthesizer_coqui"
    BARK = "synthesizer_bark"
    POLLY = "synthesizer_polly"
    CARTESIA = "synthesizer_cartesia"


class SentimentConfig(BaseModel):
    emotions: List[str] = ["angry", "friendly", "sad", "whispering"]

    @validator("emotions")
    def emotions_must_not_be_empty(cls, v):
        if len(v) == 0:
            raise ValueError("must have at least one emotion")
        return v


class SynthesizerConfig(TypedModel, type=SynthesizerType.BASE.value):  # type: ignore
    sampling_rate: int
    audio_encoding: AudioEncoding
    should_encode_as_wav: bool = False
    sentiment_config: Optional[SentimentConfig] = None

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_output_device(cls, output_device: BaseOutputDevice, **kwargs):
        return cls(
            sampling_rate=output_device.sampling_rate,
            audio_encoding=output_device.audio_encoding,
            **kwargs
        )

    # TODO(EPD-186): switch to from_twilio_output_device and from_vonage_output_device
    @classmethod
    def from_telephone_output_device(cls, **kwargs):
        return cls(
            sampling_rate=DEFAULT_SAMPLING_RATE, audio_encoding=DEFAULT_AUDIO_ENCODING, **kwargs
        )

    @classmethod
    def from_output_audio_config(cls, output_audio_config: OutputAudioConfig, **kwargs):
        return cls(
            sampling_rate=output_audio_config.sampling_rate,
            audio_encoding=output_audio_config.audio_encoding,
            **kwargs
        )


AZURE_SYNTHESIZER_DEFAULT_VOICE_NAME = "en-US-SteffanNeural"
AZURE_SYNTHESIZER_DEFAULT_PITCH = 0
AZURE_SYNTHESIZER_DEFAULT_RATE = 15


class AzureSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.AZURE.value):  # type: ignore
    voice_name: str = AZURE_SYNTHESIZER_DEFAULT_VOICE_NAME
    pitch: int = AZURE_SYNTHESIZER_DEFAULT_PITCH
    rate: int = AZURE_SYNTHESIZER_DEFAULT_RATE
    language_code: str = "en-US"


DEFAULT_GOOGLE_LANGUAGE_CODE = "en-US"
DEFAULT_GOOGLE_VOICE_NAME = "en-US-Neural2-I"
DEFAULT_GOOGLE_PITCH = 0
DEFAULT_GOOGLE_SPEAKING_RATE = 1.2


class GoogleSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.GOOGLE.value):  # type: ignore
    language_code: str = DEFAULT_GOOGLE_LANGUAGE_CODE
    voice_name: str = DEFAULT_GOOGLE_VOICE_NAME
    pitch: float = DEFAULT_GOOGLE_PITCH
    speaking_rate: float = DEFAULT_GOOGLE_SPEAKING_RATE


ELEVEN_LABS_ADAM_VOICE_ID = "pNInz6obpgDQGcFmaJgB"


class ElevenLabsSynthesizerConfig(
    SynthesizerConfig, type=SynthesizerType.ELEVEN_LABS.value  # type: ignore
):
    api_key: Optional[str] = None
    voice_id: Optional[str] = ELEVEN_LABS_ADAM_VOICE_ID
    optimize_streaming_latency: Optional[int]
    experimental_streaming: bool = False
    stability: Optional[float]
    similarity_boost: Optional[float]
    model_id: Optional[str]
    experimental_websocket: bool = False
    backchannel_amplitude_factor: float = 0.5

    @validator("voice_id")
    def set_name(cls, voice_id):
        return voice_id or ELEVEN_LABS_ADAM_VOICE_ID

    @validator("similarity_boost", always=True)
    def stability_and_similarity_boost_check(cls, similarity_boost, values):
        stability = values.get("stability")
        if (stability is None) != (similarity_boost is None):
            raise ValueError("Both stability and similarity_boost must be set or not set.")
        return similarity_boost

    @validator("optimize_streaming_latency")
    def optimize_streaming_latency_check(cls, optimize_streaming_latency):
        if optimize_streaming_latency is not None and not (0 <= optimize_streaming_latency <= 4):
            raise ValueError("optimize_streaming_latency must be between 0 and 4.")
        return optimize_streaming_latency

    @validator("backchannel_amplitude_factor")
    def backchannel_amplitude_factor_check(cls, backchannel_amplitude_factor):
        if backchannel_amplitude_factor is not None and not (0 < backchannel_amplitude_factor <= 1):
            raise ValueError(
                "backchannel_amplitude_factor must be between 0 (not inclusive) and 1."
            )
        return backchannel_amplitude_factor


RIME_DEFAULT_BASE_URL = "https://users.rime.ai/v1/rime-tts"
RIME_DEFAULT_MODEL_ID = None
RIME_DEFAULT_SPEAKER = "young_male_unmarked-1"
RIME_DEFAULT_SPEED_ALPHA = 1.0
RIME_DEFAULT_SAMPLE_RATE = SamplingRate.RATE_22050
RIME_DEFAULT_REDUCE_LATENCY = False
RimeModelId = Literal["mist", "v1"]


class RimeSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.RIME.value):  # type: ignore
    base_url: str = RIME_DEFAULT_BASE_URL
    model_id: Optional[Literal[RimeModelId]] = RIME_DEFAULT_MODEL_ID
    speaker: str = RIME_DEFAULT_SPEAKER
    speed_alpha: Optional[float] = RIME_DEFAULT_SPEED_ALPHA
    sampling_rate: int = RIME_DEFAULT_SAMPLE_RATE
    reduce_latency: Optional[bool] = RIME_DEFAULT_REDUCE_LATENCY


COQUI_DEFAULT_SPEAKER_ID = "ebe2db86-62a6-49a1-907a-9a1360d4416e"


class CoquiSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.COQUI.value):  # type: ignore
    api_key: Optional[str] = None
    voice_id: Optional[str] = COQUI_DEFAULT_SPEAKER_ID
    voice_prompt: Optional[str] = None
    use_xtts: Optional[bool] = True

    @validator("voice_id", always=True)
    def override_voice_id_with_prompt(cls, voice_id, values):
        if values.get("voice_prompt"):
            return None
        return voice_id or COQUI_DEFAULT_SPEAKER_ID


PLAYHT_DEFAULT_VOICE_ID = "larry"
PlayHtVoiceVersionType = Literal["1", "2"]


class PlayHtSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.PLAY_HT.value):  # type: ignore
    api_key: Optional[str] = None
    user_id: Optional[str] = None
    speed: Optional[float] = None
    seed: Optional[int] = None
    temperature: Optional[float] = None
    quality: Optional[str] = None
    voice_id: str = PLAYHT_DEFAULT_VOICE_ID
    experimental_streaming: bool = False
    version: Literal[PlayHtVoiceVersionType] = "2"
    top_p: Optional[float] = None
    text_guidance: Optional[float] = None
    voice_guidance: Optional[float] = None
    on_prem: bool = False
    on_prem_provider: Literal["aws", "gcp"] = "gcp"
    experimental_remove_silence: bool = False


class CoquiTTSSynthesizerConfig(
    SynthesizerConfig, type=SynthesizerType.COQUI_TTS.value  # type: ignore
):
    tts_kwargs: dict = {}
    speaker: Optional[str] = None
    language: Optional[str] = None


class GTTSSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.GTTS.value):  # type: ignore
    pass


STREAM_ELEMENTS_SYNTHESIZER_DEFAULT_VOICE = "Brian"


class StreamElementsSynthesizerConfig(
    SynthesizerConfig, type=SynthesizerType.STREAM_ELEMENTS.value  # type: ignore
):
    voice: str = STREAM_ELEMENTS_SYNTHESIZER_DEFAULT_VOICE


class BarkSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.BARK.value):  # type: ignore
    preload_kwargs: Dict[str, Any] = {}
    generate_kwargs: Dict[str, Any] = {}


DEFAULT_POLLY_LANGUAGE_CODE = "en-US"
DEFAULT_POLLY_VOICE_ID = "Matthew"
DEFAULT_POLLY_SAMPLING_RATE = SamplingRate.RATE_16000


class PollySynthesizerConfig(SynthesizerConfig, type=SynthesizerType.POLLY.value):  # type: ignore
    language_code: str = DEFAULT_POLLY_LANGUAGE_CODE
    voice_id: str = DEFAULT_POLLY_VOICE_ID
    sampling_rate: int = DEFAULT_POLLY_SAMPLING_RATE


DEFAULT_CARTESIA_MODEL_ID = 'upbeat-moon'
DEFAULT_CARTESIA_VOICE_ID = '5345cf08-6f37-424d-a5d9-8ae1101b9377'


class CartesiaSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.CARTESIA.value):  # type: ignore
    api_key: Optional[str] = None
    model_id: str = DEFAULT_CARTESIA_MODEL_ID
    voice_id: str = DEFAULT_CARTESIA_VOICE_ID



from typing import Type
from vocode.streaming.models.synthesizer import (
    AzureSynthesizerConfig,
    CartesiaSynthesizerConfig,
    ElevenLabsSynthesizerConfig,
    PlayHtSynthesizerConfig,
    RimeSynthesizerConfig,
    StreamElementsSynthesizerConfig,
    SynthesizerConfig,
)
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.synthesizer.azure_synthesizer import AzureSynthesizer
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer
from vocode.streaming.synthesizer.cartesia_synthesizer import CartesiaSynthesizer
from vocode.streaming.synthesizer.eleven_labs_synthesizer import ElevenLabsSynthesizer
from vocode.streaming.synthesizer.eleven_labs_websocket_synthesizer import ElevenLabsWSSynthesizer
from vocode.streaming.synthesizer.play_ht_synthesizer import PlayHtSynthesizer
from vocode.streaming.synthesizer.play_ht_synthesizer_v2 import PlayHtSynthesizerV2
from vocode.streaming.synthesizer.rime_synthesizer import RimeSynthesizer
from vocode.streaming.synthesizer.stream_elements_synthesizer import StreamElementsSynthesizer


class DefaultSynthesizerFactory(AbstractSynthesizerFactory):
    def create_synthesizer(
        self,
        synthesizer_config: SynthesizerConfig,
    ):
        if isinstance(synthesizer_config, AzureSynthesizerConfig):
            return AzureSynthesizer(synthesizer_config)
        elif isinstance(synthesizer_config, CartesiaSynthesizerConfig):
            return CartesiaSynthesizer(synthesizer_config)
        elif isinstance(synthesizer_config, ElevenLabsSynthesizerConfig):
            eleven_labs_synthesizer_class_type: Type[BaseSynthesizer] = ElevenLabsSynthesizer
            if synthesizer_config.experimental_websocket:
                eleven_labs_synthesizer_class_type = ElevenLabsWSSynthesizer
            return eleven_labs_synthesizer_class_type(synthesizer_config)
        elif isinstance(synthesizer_config, PlayHtSynthesizerConfig):
            if synthesizer_config.version == "2":
                return PlayHtSynthesizerV2(synthesizer_config)
            else:
                return PlayHtSynthesizer(synthesizer_config)
        elif isinstance(synthesizer_config, RimeSynthesizerConfig):
            return RimeSynthesizer(synthesizer_config)
        elif isinstance(synthesizer_config, StreamElementsSynthesizerConfig):
            return StreamElementsSynthesizer(synthesizer_config)
        else:
            raise Exception("Invalid synthesizer config")



import os
from typing import Optional, TypeVar

from loguru import logger
from redis.asyncio import Redis
from redis.backoff import ExponentialBackoff, NoBackoff
from redis.exceptions import ConnectionError, TimeoutError
from redis.retry import Retry

from vocode.streaming.utils.singleton import Singleton

WorkerInputType = TypeVar("WorkerInputType")
# Two separate factories for Redis clients so that the
# typing gets picked up properly


def initialize_redis(retries: int = 1):
    backoff = ExponentialBackoff() if retries > 1 else NoBackoff()
    retry = Retry(backoff, retries)
    return Redis(  # type: ignore
        host=os.environ.get("REDISHOST", "localhost"),
        port=int(os.environ.get("REDISPORT", 6379)),
        username=os.environ.get("REDISUSER", None),
        password=os.environ.get("REDISPASSWORD", None),
        decode_responses=True,
        retry=retry,
        ssl=bool(os.environ.get("REDISSSL", False)),
        ssl_cert_reqs="none",
        retry_on_error=[ConnectionError, TimeoutError],
        health_check_interval=30,
    )


def initialize_redis_bytes():
    return Redis(
        host=os.environ.get("REDISHOST", "localhost"),
        port=int(os.environ.get("REDISPORT", 6379)),
        username=os.environ.get("REDISUSER", None),
        password=os.environ.get("REDISPASSWORD", None),
        db=0,
        ssl=bool(os.environ.get("REDISSSL", False)),
        ssl_cert_reqs="none",
    )


class RedisGenericMessageQueue(Singleton):
    """Class representing a Redis Streams message queue."""

    def __init__(
        self,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> None:
        """
        Initialize a RedisGenericMessageQueue instance.

        This initializes a Redis client and sets the name of the stream.
        """
        self.redis: Redis = initialize_redis()
        queue_name_prefix = f"{prefix}_" if prefix else ""
        queue_name_suffix = f"_{suffix}" if suffix else ""
        self.queue_name = f"{queue_name_prefix}queue{queue_name_suffix}"

    async def publish(self, message: dict) -> None:
        """
        Publishes a message to the Redis stream.

        Args:
            message (dict): The message to be published.

        Returns:
            None
        """
        logger.info(f"[{self.queue_name}] Publishing message: {message}")
        try:
            await self.redis.xadd(self.queue_name, message)
        except Exception as e:
            logger.exception(f"[{self.queue_name}] Failed to publish message: {message}")
            raise e



from typing import AsyncGenerator, Tuple

from vocode.streaming.agent.base_agent import GeneratedResponse, RespondAgent
from vocode.streaming.models.agent import EchoAgentConfig
from vocode.streaming.models.message import BaseMessage


class EchoAgent(RespondAgent[EchoAgentConfig]):
    async def respond(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
    ) -> Tuple[str, bool]:
        return human_input, False

    async def generate_response(
        self,
        human_input,
        conversation_id: str,
        is_interrupt: bool = False,
        bot_was_in_medias_res: bool = False,
    ) -> AsyncGenerator[GeneratedResponse, None]:
        yield GeneratedResponse(message=BaseMessage(text=human_input), is_interruptible=True)

    def update_last_bot_message_on_cut_off(self, message: str):
        pass



from abc import ABC, abstractmethod

from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.transcriber.base_transcriber import (
    BaseAsyncTranscriber,
    BaseThreadAsyncTranscriber,
)


class AbstractTranscriberFactory(ABC):
    @abstractmethod
    def create_transcriber(
        self,
        transcriber_config: TranscriberConfig,
    ) -> BaseAsyncTranscriber[TranscriberConfig] | BaseThreadAsyncTranscriber[TranscriberConfig]:
        pass



import json
from typing import Any, Dict, Optional, Type

from pydantic.v1 import BaseModel

from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.action.external_actions_requester import (
    ExternalActionResponse,
    ExternalActionsRequester,
)
from vocode.streaming.models.actions import ActionConfig as VocodeActionConfig
from vocode.streaming.models.actions import ActionInput, ActionOutput, ExternalActionProcessingMode
from vocode.streaming.models.message import BaseMessage


class ExecuteExternalActionVocodeActionConfig(
    VocodeActionConfig, type="action_external"  # type: ignore
):
    processing_mode: ExternalActionProcessingMode
    name: str
    description: str
    url: str
    input_schema: str
    speak_on_send: bool
    speak_on_receive: bool
    signature_secret: str


class ExecuteExternalActionParameters(BaseModel):
    payload: Dict[str, Any]


class ExecuteExternalActionResponse(BaseModel):
    success: bool
    result: Optional[dict]


class ExecuteExternalAction(
    BaseAction[
        ExecuteExternalActionVocodeActionConfig,
        ExecuteExternalActionParameters,
        ExecuteExternalActionResponse,
    ]
):
    parameters_type: Type[ExecuteExternalActionParameters] = ExecuteExternalActionParameters
    response_type: Type[ExecuteExternalActionResponse] = ExecuteExternalActionResponse

    def __init__(
        self,
        action_config: ExecuteExternalActionVocodeActionConfig,
    ):
        self.description = action_config.description
        super().__init__(
            action_config,
            quiet=not action_config.speak_on_receive,
            should_respond="always" if action_config.speak_on_send else "never",
            is_interruptible=False,
        )
        self.external_actions_requester = ExternalActionsRequester(url=action_config.url)

    def _user_message_param_info(self):
        return {
            "type": "string",
            "description": (
                "A message to reply to the user with BEFORE we make the function call."
                "\nEssentially a live response informing them that the function is "
                "about to happen."
            ),
        }

    def create_action_params(self, params: Dict[str, Any]) -> ExecuteExternalActionParameters:
        return ExecuteExternalActionParameters(payload=params)

    def get_function_name(self) -> str:
        return self.action_config.name

    def get_parameters_schema(self) -> Dict[str, Any]:
        return json.loads(self.action_config.input_schema)

    async def send_external_action_request(
        self, action_input: ActionInput[ExecuteExternalActionParameters]
    ) -> ExternalActionResponse:
        return await self.external_actions_requester.send_request(
            payload=action_input.params.payload,
            signature_secret=self.action_config.signature_secret,
        )

    async def run(
        self, action_input: ActionInput[ExecuteExternalActionParameters]
    ) -> ActionOutput[ExecuteExternalActionResponse]:
        # TODO: this interruption handling needs to be refactored / DRYd
        if self.should_respond and action_input.user_message_tracker is not None:
            await action_input.user_message_tracker.wait()

        self.conversation_state_manager.mute_agent()
        response = await self.send_external_action_request(action_input)
        self.conversation_state_manager.unmute_agent()

        # TODO (EA): pass specific context based on error
        return ActionOutput(
            action_type=action_input.action_config.type,
            response=ExecuteExternalActionResponse(
                result=response.result if response else None, success=response.success
            ),
            canned_response=(
                BaseMessage(text=response.agent_message)
                if response and response.agent_message
                else None
            ),
        )



from vocode.streaming.models.message import LLMToken


class InputStreamingSynthesizer:
    async def handle_end_of_turn(self):
        pass

    def get_current_utterance_synthesis_result(self):
        raise NotImplementedError

    async def send_token_to_synthesizer(
        self,
        message: LLMToken,
        chunk_size: int,
    ):
        raise NotImplementedError



import re
from typing import Optional, Type

from pydantic.v1 import BaseModel, Field

from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.models.actions import ActionConfig, ActionInput, ActionOutput

EMAIL_REGEX = r"^(?!\.)(?!.*\.\.)[a-zA-Z0-9._%+-]+(?<!\.)@(?![.])[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"


class RecordEmailVocodeActionConfig(ActionConfig, type="action_record_email"):  # type: ignore
    pass


class RecordEmailParameters(BaseModel):
    descriptor: str = Field("A human readable descriptor; eg 'The user's email.'")
    raw_value: str = Field(
        ...,
        description="The raw value parsed from the transcript.",
    )

    formatted_value: str = Field(
        ...,
        description="The estimated FORMATTED value of the email.",
    )


class RecordEmailResponse(BaseModel):
    success: bool
    message: Optional[str] = None


class RecordEmail(
    BaseAction[
        RecordEmailVocodeActionConfig,
        RecordEmailParameters,
        RecordEmailResponse,
    ]
):
    description: str = """Attempts to record an email from the transcript.

    You must format the value to match the field type, eg

    kian at g mail dot com -> kian@gmail.com
    ajay at vocode dot dev -> ajay@vocode.dev

    This function will do extra validation.
    """
    parameters_type: Type[RecordEmailParameters] = RecordEmailParameters
    response_type: Type[RecordEmailResponse] = RecordEmailResponse

    def __init__(
        self,
        action_config: RecordEmailVocodeActionConfig,
    ):
        super().__init__(
            action_config,
            quiet=False,
            should_respond="never",
        )

    def _validate_email(self, email: str) -> bool:
        return bool(re.match(EMAIL_REGEX, email))

    async def run(
        self, action_input: ActionInput[RecordEmailParameters]
    ) -> ActionOutput[RecordEmailResponse]:
        value = action_input.params.formatted_value

        success = self._validate_email(value)

        return ActionOutput(
            action_type=action_input.action_config.type,
            response=RecordEmailResponse(success=success),
        )



TEXT_TO_SPEECH_CHUNK_SIZE_SECONDS = 0.1
PER_CHUNK_ALLOWANCE_SECONDS = 0.01
ALLOWED_IDLE_TIME = 15
SENTENCE_ENDINGS = [".", "!", "?", "\n"]
CHECK_HUMAN_PRESENT_MESSAGE_CHOICES = [
    "Hello?",
    "Are you there?",
    "Are you still there?",
    "Hi, are you there?",
]



from typing import Dict, Sequence, Type

from vocode.streaming.action.abstract_factory import AbstractActionFactory
from vocode.streaming.action.base_action import BaseAction
from vocode.streaming.action.dtmf import TwilioDTMF, VonageDTMF
from vocode.streaming.action.end_conversation import EndConversation
from vocode.streaming.action.execute_external_action import ExecuteExternalAction
from vocode.streaming.action.phone_call_action import (
    TwilioPhoneConversationAction,
    VonagePhoneConversationAction,
)
from vocode.streaming.action.record_email import RecordEmail
from vocode.streaming.action.transfer_call import TwilioTransferCall, VonageTransferCall
from vocode.streaming.action.wait import Wait
from vocode.streaming.models.actions import ActionConfig, ActionType

CONVERSATION_ACTIONS: Dict[ActionType, Type[BaseAction]] = {
    ActionType.END_CONVERSATION: EndConversation,
    ActionType.RECORD_EMAIL: RecordEmail,
    ActionType.WAIT: Wait,
    ActionType.EXECUTE_EXTERNAL_ACTION: ExecuteExternalAction,
}

VONAGE_ACTIONS: Dict[ActionType, Type[VonagePhoneConversationAction]] = {
    ActionType.TRANSFER_CALL: VonageTransferCall,
    ActionType.DTMF: VonageDTMF,
}

TWILIO_ACTIONS: Dict[ActionType, Type[TwilioPhoneConversationAction]] = {
    ActionType.TRANSFER_CALL: TwilioTransferCall,
    ActionType.DTMF: TwilioDTMF,
}


class DefaultActionFactory(AbstractActionFactory):
    def __init__(self, actions: Sequence[ActionConfig] | dict = {}):

        self.action_configs_dict = {action.type: action for action in actions}
        self.actions = CONVERSATION_ACTIONS

    def create_action(self, action_config: ActionConfig):
        if action_config.type not in self.action_configs_dict:
            raise Exception("Action type not supported by Agent config.")

        action_class = self.actions[action_config.type]

        return action_class(action_config)



import io
import wave

import miniaudio


# sampling_rate is the rate of the input, not expected output
def decode_mp3(mp3_bytes: bytes) -> io.BytesIO:
    # Convert it to a wav chunk using miniaudio
    wav_chunk = miniaudio.decode(mp3_bytes, nchannels=1)

    # Write wav_chunks.samples to io.BytesIO with builtin WAVE
    output_bytes_io = io.BytesIO()

    with wave.open(output_bytes_io, "wb") as wave_obj:
        wave_obj.setnchannels(1)
        wave_obj.setsampwidth(2)
        wave_obj.setframerate(44100)
        wave_obj.writeframes(wav_chunk.samples)
    output_bytes_io.seek(0)
    return output_bytes_io
